# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Helpers functions for different kinds of WCSAxes instances.
"""

import numpy as np
from matplotlib.offsetbox import AnchoredOffsetbox, AuxTransformBox
from matplotlib.patches import Ellipse
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar

import astropy.units as u
from astropy.wcs.utils import proj_plane_pixel_scales

__all__ = ["add_beam", "add_scalebar"]

CORNERS = {
    "top right": 1,
    "top left": 2,
    "bottom left": 3,
    "bottom right": 4,
    "right": 5,
    "left": 6,
    "bottom": 8,
    "top": 9,
}


def add_beam(
    ax,
    header=None,
    major=None,
    minor=None,
    angle=None,
    corner="bottom left",
    frame=False,
    borderpad=0.4,
    pad=0.5,
    **kwargs,
):
    """
    Display the beam shape and size.

    Parameters
    ----------
    ax : :class:`~astropy.visualization.wcsaxes.WCSAxes`
        WCSAxes instance in which the beam shape and size is displayed. The WCS
        must be celestial.
    header : :class:`~astropy.io.fits.Header`, optional
        Header containing the beam parameters. If specified, the ``BMAJ``,
        ``BMIN``, and ``BPA`` keywords will be searched in the FITS header
        to set the major and minor axes and the position angle on the sky.
    major : float or :class:`~astropy.units.Quantity`, optional
        Major axis of the beam in degrees or an angular quantity.
    minor : float, or :class:`~astropy.units.Quantity`, optional
        Minor axis of the beam in degrees or an angular quantity.
    angle : float or :class:`~astropy.units.Quantity`, optional
        Position angle of the beam on the sky in degrees or an angular
        quantity in the anticlockwise direction.
    corner : str, optional
        The beam location. Acceptable values are ``'left'``, ``'right'``,
        ``'top'``, 'bottom', ``'top left'``, ``'top right'``, ``'bottom left'``
        (default), and ``'bottom right'``.
    frame : bool, optional
        Whether to display a frame behind the beam (default is ``False``).
    borderpad : float, optional
        Border padding, in fraction of the font size. Default is 0.4.
    pad : float, optional
        Padding around the beam, in fraction of the font size. Default is 0.5.
    kwargs
        Additional arguments are passed to :class:`matplotlib.patches.Ellipse`.

    Notes
    -----
    This function may be inaccurate when:

    - The pixel scales at the reference pixel are different from the pixel scales
      within the image extent (e.g., when the reference pixel is well outside of
      the image extent and the projection is non-linear)
    - The pixel scales in the two directions are very different from each other
      (e.g., rectangular pixels)

    """
    if header and major:
        raise ValueError(
            "Either header or major/minor/angle must be specified, not both."
        )

    if header:
        major = header["BMAJ"]
        minor = header["BMIN"]
        angle = header["BPA"]

    if isinstance(major, u.Quantity):
        major = major.to(u.degree).value

    if isinstance(minor, u.Quantity):
        minor = minor.to(u.degree).value

    if isinstance(angle, u.Quantity):
        angle = angle.to(u.degree).value

    if ax.wcs.is_celestial:
        pix_scale = proj_plane_pixel_scales(ax.wcs)
        sx = pix_scale[0]
        sy = pix_scale[1]
        degrees_per_pixel = np.sqrt(sx * sy)
    else:
        raise ValueError("Cannot show beam when WCS is not celestial")

    minor /= degrees_per_pixel
    major /= degrees_per_pixel

    aux_tr_box = AuxTransformBox(ax.transData)
    ellipse = Ellipse((0, 0), width=minor, height=major, angle=angle, **kwargs)
    aux_tr_box.add_artist(ellipse)
    box = AnchoredOffsetbox(
        child=aux_tr_box,
        pad=pad,
        borderpad=borderpad,
        loc=CORNERS[corner],
        frameon=frame,
    )
    ax.add_artist(box)


def add_scalebar(
    ax,
    length,
    label=None,
    corner="bottom right",
    frame=False,
    borderpad=0.4,
    pad=0.5,
    **kwargs,
):
    """Add a scale bar.

    Parameters
    ----------
    ax : :class:`~astropy.visualization.wcsaxes.WCSAxes`
        WCSAxes instance in which the scale bar is displayed. The WCS must be
        celestial.
    length : float or :class:`~astropy.units.Quantity`
        The length of the scalebar in degrees or an angular quantity
    label : str, optional
        Label to place below the scale bar
    corner : str, optional
        Where to place the scale bar. Acceptable values are:, ``'left'``,
        ``'right'``, ``'top'``, ``'bottom'``, ``'top left'``, ``'top right'``,
        ``'bottom left'`` and ``'bottom right'`` (default)
    frame : bool, optional
        Whether to display a frame behind the scale bar (default is ``False``)
    borderpad : float, optional
        Border padding, in fraction of the font size. Default is 0.4.
    pad : float, optional
        Padding around the scale bar, in fraction of the font size. Default is 0.5.
    kwargs
        Additional arguments are passed to
        :class:`mpl_toolkits.axes_grid1.anchored_artists.AnchoredSizeBar`.

    Notes
    -----
    This function may be inaccurate when:

    - The pixel scales at the reference pixel are different from the pixel scales
      within the image extent (e.g., when the reference pixel is well outside of
      the image extent and the projection is non-linear)
    - The pixel scales in the two directions are very different from each other
      (e.g., rectangular pixels)

    """
    if isinstance(length, u.Quantity):
        length = length.to(u.degree).value

    if ax.wcs.is_celestial:
        pix_scale = proj_plane_pixel_scales(ax.wcs)
        sx = pix_scale[0]
        sy = pix_scale[1]
        degrees_per_pixel = np.sqrt(sx * sy)
    else:
        raise ValueError("Cannot show scalebar when WCS is not celestial")

    length = length / degrees_per_pixel

    corner = CORNERS[corner]

    scalebar = AnchoredSizeBar(
        ax.transData,
        length,
        label,
        corner,
        pad=pad,
        borderpad=borderpad,
        sep=5,
        frameon=frame,
        **kwargs,
    )

    ax.add_artist(scalebar)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# ruff: noqa: RUF009


import numpy as np
from numpy import sqrt

from astropy.cosmology._utils import aszarr
from astropy.cosmology.core import dataclass_decorator
from astropy.cosmology.parameter import Parameter

from . import scalar_inv_efuncs
from .base import FLRW, FlatFLRWMixin

__all__ = ["wCDM", "FlatwCDM"]

__doctest_requires__ = {"*": ["scipy"]}


@dataclass_decorator
class wCDM(FLRW):
    """FLRW cosmology with a constant dark energy EoS and curvature.

    This has one additional attribute beyond those of FLRW.

    Parameters
    ----------
    H0 : float or scalar quantity-like ['frequency']
        Hubble constant at z = 0. If a float, must be in [km/sec/Mpc].

    Om0 : float
        Omega matter: density of non-relativistic matter in units of the
        critical density at z=0.

    Ode0 : float
        Omega dark energy: density of dark energy in units of the critical
        density at z=0.

    w0 : float, optional
        Dark energy equation of state at all redshifts. This is
        pressure/density for dark energy in units where c=1. A cosmological
        constant has w0=-1.0.

    Tcmb0 : float or scalar quantity-like ['temperature'], optional
        Temperature of the CMB z=0. If a float, must be in [K]. Default: 0 [K].
        Setting this to zero will turn off both photons and neutrinos
        (even massive ones).

    Neff : float, optional
        Effective number of Neutrino species. Default 3.04.

    m_nu : quantity-like ['energy', 'mass'] or array-like, optional
        Mass of each neutrino species in [eV] (mass-energy equivalency enabled).
        If this is a scalar Quantity, then all neutrino species are assumed to
        have that mass. Otherwise, the mass of each species. The actual number
        of neutrino species (and hence the number of elements of m_nu if it is
        not scalar) must be the floor of Neff. Typically this means you should
        provide three neutrino masses unless you are considering something like
        a sterile neutrino.

    Ob0 : float or None, optional
        Omega baryons: density of baryonic matter in units of the critical
        density at z=0.  If this is set to None (the default), any computation
        that requires its value will raise an exception.

    name : str or None (optional, keyword-only)
        Name for this cosmological object.

    meta : mapping or None (optional, keyword-only)
        Metadata for the cosmology, e.g., a reference.

    Examples
    --------
    >>> from astropy.cosmology import wCDM
    >>> cosmo = wCDM(H0=70, Om0=0.3, Ode0=0.7, w0=-0.9)

    The comoving distance in Mpc at redshift z:

    >>> z = 0.5
    >>> dc = cosmo.comoving_distance(z)
    """

    w0: Parameter = Parameter(
        default=-1.0, doc="Dark energy equation of state.", fvalidate="float"
    )

    def __post_init__(self):
        super().__post_init__()

        # Please see :ref:`astropy-cosmology-fast-integrals` for discussion
        # about what is being done here.
        if self._Tcmb0.value == 0:
            inv_efunc_scalar = scalar_inv_efuncs.wcdm_inv_efunc_norel
            inv_efunc_scalar_args = (self._Om0, self._Ode0, self._Ok0, self._w0)
        elif not self._massivenu:
            inv_efunc_scalar = scalar_inv_efuncs.wcdm_inv_efunc_nomnu
            inv_efunc_scalar_args = (
                self._Om0,
                self._Ode0,
                self._Ok0,
                self._Ogamma0 + self._Onu0,
                self._w0,
            )
        else:
            inv_efunc_scalar = scalar_inv_efuncs.wcdm_inv_efunc
            inv_efunc_scalar_args = (
                self._Om0,
                self._Ode0,
                self._Ok0,
                self._Ogamma0,
                self._neff_per_nu,
                self._nmasslessnu,
                self._nu_y_list,
                self._w0,
            )

        object.__setattr__(self, "_inv_efunc_scalar", inv_efunc_scalar)
        object.__setattr__(self, "_inv_efunc_scalar_args", inv_efunc_scalar_args)

    def w(self, z):
        r"""Returns dark energy equation of state at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, or `~numbers.Number`
            Input redshift.

        Returns
        -------
        w : ndarray or float
            The dark energy equation of state
            Returns `float` if the input is scalar.

        Notes
        -----
        The dark energy equation of state is defined as
        :math:`w(z) = P(z)/\rho(z)`, where :math:`P(z)` is the pressure at
        redshift z and :math:`\rho(z)` is the density at redshift z, both in
        units where c=1. Here this is :math:`w(z) = w_0`.
        """
        z = aszarr(z)
        return self._w0 * (np.ones(z.shape) if hasattr(z, "shape") else 1.0)

    def de_density_scale(self, z):
        r"""Evaluates the redshift dependence of the dark energy density.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, or `~numbers.Number`
            Input redshift.

        Returns
        -------
        I : ndarray or float
            The scaling of the energy density of dark energy with redshift.
            Returns `float` if the input is scalar.

        Notes
        -----
        The scaling factor, I, is defined by :math:`\rho(z) = \rho_0 I`,
        and in this case is given by
        :math:`I = \left(1 + z\right)^{3\left(1 + w_0\right)}`
        """
        return (aszarr(z) + 1.0) ** (3.0 * (1.0 + self._w0))

    def efunc(self, z):
        """Function used to calculate H(z), the Hubble parameter.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, or `~numbers.Number`
            Input redshift.

        Returns
        -------
        E : ndarray or float
            The redshift scaling of the Hubble constant.
            Returns `float` if the input is scalar.
            Defined such that :math:`H(z) = H_0 E(z)`.
        """
        Or = self._Ogamma0 + (
            self._Onu0
            if not self._massivenu
            else self._Ogamma0 * self.nu_relative_density(z)
        )
        zp1 = aszarr(z) + 1.0  # (converts z [unit] -> z [dimensionless])

        return sqrt(
            zp1**2 * ((Or * zp1 + self._Om0) * zp1 + self._Ok0)
            + self._Ode0 * zp1 ** (3.0 * (1.0 + self._w0))
        )

    def inv_efunc(self, z):
        r"""Function used to calculate :math:`\frac{1}{H_z}`.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, or `~numbers.Number`
            Input redshift.

        Returns
        -------
        E : ndarray or float
            The inverse redshift scaling of the Hubble constant.
            Returns `float` if the input is scalar.
            Defined such that :math:`H_z = H_0 / E`.
        """
        Or = self._Ogamma0 + (
            self._Onu0
            if not self._massivenu
            else self._Ogamma0 * self.nu_relative_density(z)
        )
        zp1 = aszarr(z) + 1.0  # (converts z [unit] -> z [dimensionless])

        return (
            zp1**2 * ((Or * zp1 + self._Om0) * zp1 + self._Ok0)
            + self._Ode0 * zp1 ** (3.0 * (1.0 + self._w0))
        ) ** (-0.5)


@dataclass_decorator
class FlatwCDM(FlatFLRWMixin, wCDM):
    """FLRW cosmology with a constant dark energy EoS and no spatial curvature.

    This has one additional attribute beyond those of FLRW.

    Parameters
    ----------
    H0 : float or scalar quantity-like ['frequency']
        Hubble constant at z = 0. If a float, must be in [km/sec/Mpc].

    Om0 : float
        Omega matter: density of non-relativistic matter in units of the
        critical density at z=0.

    w0 : float, optional
        Dark energy equation of state at all redshifts. This is
        pressure/density for dark energy in units where c=1. A cosmological
        constant has w0=-1.0.

    Tcmb0 : float or scalar quantity-like ['temperature'], optional
        Temperature of the CMB z=0. If a float, must be in [K]. Default: 0 [K].
        Setting this to zero will turn off both photons and neutrinos
        (even massive ones).

    Neff : float, optional
        Effective number of Neutrino species. Default 3.04.

    m_nu : quantity-like ['energy', 'mass'] or array-like, optional
        Mass of each neutrino species in [eV] (mass-energy equivalency enabled).
        If this is a scalar Quantity, then all neutrino species are assumed to
        have that mass. Otherwise, the mass of each species. The actual number
        of neutrino species (and hence the number of elements of m_nu if it is
        not scalar) must be the floor of Neff. Typically this means you should
        provide three neutrino masses unless you are considering something like
        a sterile neutrino.

    Ob0 : float or None, optional
        Omega baryons: density of baryonic matter in units of the critical
        density at z=0.  If this is set to None (the default), any computation
        that requires its value will raise an exception.

    name : str or None (optional, keyword-only)
        Name for this cosmological object.

    meta : mapping or None (optional, keyword-only)
        Metadata for the cosmology, e.g., a reference.

    Examples
    --------
    >>> from astropy.cosmology import FlatwCDM
    >>> cosmo = FlatwCDM(H0=70, Om0=0.3, w0=-0.9)

    The comoving distance in Mpc at redshift z:

    >>> z = 0.5
    >>> dc = cosmo.comoving_distance(z)

    To get an equivalent cosmology, but of type `astropy.cosmology.wCDM`,
    use :attr:`astropy.cosmology.FlatFLRWMixin.nonflat`.

    >>> print(cosmo.nonflat)
    wCDM(H0=70.0 km / (Mpc s), Om0=0.3, Ode0=0.7, ...
    """

    def __post_init__(self):
        super().__post_init__()

        # Please see :ref:`astropy-cosmology-fast-integrals` for discussion
        # about what is being done here.
        if self._Tcmb0.value == 0:
            inv_efunc_scalar = scalar_inv_efuncs.fwcdm_inv_efunc_norel
            inv_efunc_scalar_args = (self._Om0, self._Ode0, self._w0)
        elif not self._massivenu:
            inv_efunc_scalar = scalar_inv_efuncs.fwcdm_inv_efunc_nomnu
            inv_efunc_scalar_args = (
                self._Om0,
                self._Ode0,
                self._Ogamma0 + self._Onu0,
                self._w0,
            )
        else:
            inv_efunc_scalar = scalar_inv_efuncs.fwcdm_inv_efunc
            inv_efunc_scalar_args = (
                self._Om0,
                self._Ode0,
                self._Ogamma0,
                self._neff_per_nu,
                self._nmasslessnu,
                self._nu_y_list,
                self._w0,
            )
        object.__setattr__(self, "_inv_efunc_scalar", inv_efunc_scalar)
        object.__setattr__(self, "_inv_efunc_scalar_args", inv_efunc_scalar_args)

    def efunc(self, z):
        """Function used to calculate H(z), the Hubble parameter.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, or `~numbers.Number`
            Input redshift.

        Returns
        -------
        E : ndarray or float
            The redshift scaling of the Hubble constant.
            Returns `float` if the input is scalar.
            Defined such that :math:`H(z) = H_0 E(z)`.
        """
        Or = self._Ogamma0 + (
            self._Onu0
            if not self._massivenu
            else self._Ogamma0 * self.nu_relative_density(z)
        )
        zp1 = aszarr(z) + 1.0  # (converts z [unit] -> z [dimensionless])

        return sqrt(
            zp1**3 * (Or * zp1 + self._Om0) + self._Ode0 * zp1 ** (3.0 * (1 + self._w0))
        )

    def inv_efunc(self, z):
        r"""Function used to calculate :math:`\frac{1}{H_z}`.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, or `~numbers.Number`
            Input redshift.

        Returns
        -------
        E : ndarray or float
            The inverse redshift scaling of the Hubble constant.
            Returns `float` if the input is scalar.
            Defined such that :math:`H(z) = H_0 E(z)`.
        """
        Or = self._Ogamma0 + (
            self._Onu0
            if not self._massivenu
            else self._Ogamma0 * self.nu_relative_density(z)
        )
        zp1 = aszarr(z) + 1.0  # (converts z [unit] -> z [dimensionless])

        return (
            zp1**3 * (Or * zp1 + self._Om0)
            + self._Ode0 * zp1 ** (3.0 * (1.0 + self._w0))
        ) ** (-0.5)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""|Cosmology| I/O, using |Cosmology.to_format| and |Cosmology.from_format|.

This module provides functions to transform a |Cosmology| object to and from another
|Cosmology| object. The functions are registered with ``convert_registry`` under the
format name "astropy.cosmology". You probably won't need to use these functions as they
are present mainly for completeness and testing.

    >>> from astropy.cosmology import Cosmology, Planck18
    >>> Planck18.to_format("astropy.cosmology") is Planck18
    True
    >>> Cosmology.from_format(Planck18) is Planck18
    True
"""

from astropy.cosmology.connect import convert_registry
from astropy.cosmology.core import _COSMOLOGY_CLASSES, Cosmology

__all__ = []  # nothing is publicly scoped


def from_cosmology(cosmo, /, **kwargs):
    """Return the |Cosmology| unchanged.

    Parameters
    ----------
    cosmo : `~astropy.cosmology.Cosmology`
        The cosmology to return.
    **kwargs
        This argument is required for compatibility with the standard set of
        keyword arguments in format `~astropy.cosmology.Cosmology.from_format`,
        e.g. "cosmology". If "cosmology" is included and is not `None`,
        ``cosmo`` is checked for correctness.

    Returns
    -------
    `~astropy.cosmology.Cosmology` subclass instance
        Just ``cosmo`` passed through.

    Raises
    ------
    TypeError
        If the |Cosmology| object is not an instance of ``cosmo`` (and
        ``cosmology`` is not `None`).

    Examples
    --------
    >>> from astropy.cosmology import Cosmology, Planck18
    >>> print(Cosmology.from_format(Planck18))
    FlatLambdaCDM(name="Planck18", H0=67.66 km / (Mpc s), Om0=0.30966,
                  Tcmb0=2.7255 K, Neff=3.046, m_nu=[0. 0. 0.06] eV, Ob0=0.04897)
    """
    # Check argument `cosmology`
    cosmology = kwargs.get("cosmology")
    if isinstance(cosmology, str):
        cosmology = _COSMOLOGY_CLASSES[cosmology]
    if cosmology is not None and not isinstance(cosmo, cosmology):
        raise TypeError(f"cosmology {cosmo} is not an {cosmology} instance.")

    return cosmo


def to_cosmology(cosmo, *args):
    """Return the |Cosmology| unchanged.

    Parameters
    ----------
    cosmo : `~astropy.cosmology.Cosmology`
        The cosmology to return.
    *args
        Not used.

    Returns
    -------
    `~astropy.cosmology.Cosmology` subclass instance
        Just ``cosmo`` passed through.

    Examples
    --------
    >>> from astropy.cosmology import Planck18
    >>> Planck18.to_format("astropy.cosmology") is Planck18
    True
    """
    return cosmo


def cosmology_identify(origin, format, *args, **kwargs):
    """Identify if object is a `~astropy.cosmology.Cosmology`.

    Returns
    -------
    bool
    """
    itis = False
    if origin == "read":
        itis = isinstance(args[1], Cosmology) and (
            format in (None, "astropy.cosmology")
        )
    return itis


# ===================================================================
# Register

convert_registry.register_reader("astropy.cosmology", Cosmology, from_cosmology)
convert_registry.register_writer("astropy.cosmology", Cosmology, to_cosmology)
convert_registry.register_identifier("astropy.cosmology", Cosmology, cosmology_identify)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This is the APE5 coordinates API document re-written to work as a series of test
functions.

Note that new tests for coordinates functionality should generally *not* be
added to this file - instead, add them to other appropriate test modules  in
this package, like ``test_sky_coord.py``, ``test_frames.py``, or
``test_representation.py``.  This file is instead meant mainly to keep track of
deviations from the original APE5 plan.
"""

import numpy as np
import pytest
from numpy import testing as npt

from astropy import coordinates as coords
from astropy import time
from astropy import units as u
from astropy.tests.helper import assert_quantity_allclose as assert_allclose
from astropy.units import allclose
from astropy.utils.compat.optional_deps import HAS_SCIPY


def test_representations_api():
    from astropy.coordinates import Angle, Distance, Latitude, Longitude
    from astropy.coordinates.representation import (
        CartesianRepresentation,
        PhysicsSphericalRepresentation,
        SphericalRepresentation,
        UnitSphericalRepresentation,
    )

    # <-----------------Classes for representation of coordinate data-------------->
    # These classes inherit from a common base class and internally contain Quantity
    # objects, which are arrays (although they may act as scalars, like numpy's
    # length-0  "arrays")
    # They can be initialized with a variety of ways that make intuitive sense.
    # Distance is optional.
    UnitSphericalRepresentation(lon=8 * u.hour, lat=5 * u.deg)
    UnitSphericalRepresentation(lon=8 * u.hourangle, lat=5 * u.deg)
    SphericalRepresentation(lon=8 * u.hourangle, lat=5 * u.deg, distance=10 * u.kpc)

    # In the initial implementation, the lat/lon/distance arguments to the
    # initializer must be in order. A *possible* future change will be to allow
    # smarter guessing of the order.  E.g. `Latitude` and `Longitude` objects can be
    # given in any order.
    UnitSphericalRepresentation(Longitude(8, u.hour), Latitude(5, u.deg))
    SphericalRepresentation(
        Longitude(8, u.hour), Latitude(5, u.deg), Distance(10, u.kpc)
    )

    # Arrays of any of the inputs are fine
    UnitSphericalRepresentation(lon=[8, 9] * u.hourangle, lat=[5, 6] * u.deg)

    # Default is to copy arrays, but optionally, it can be a reference
    UnitSphericalRepresentation(
        lon=[8, 9] * u.hourangle, lat=[5, 6] * u.deg, copy=False
    )

    # strings are parsed by `Latitude` and `Longitude` constructors, so no need to
    # implement parsing in the Representation classes
    UnitSphericalRepresentation(lon=Angle("2h6m3.3s"), lat=Angle("0.1rad"))

    # Or, you can give `Quantity`s with keywords, and they will be internally
    # converted to Angle/Distance
    c1 = SphericalRepresentation(
        lon=8 * u.hourangle, lat=5 * u.deg, distance=10 * u.kpc
    )

    # Can also give another representation object with the `reprobj` keyword.
    c2 = SphericalRepresentation.from_representation(c1)

    #  distance, lat, and lon typically will just match in shape
    SphericalRepresentation(
        lon=[8, 9] * u.hourangle, lat=[5, 6] * u.deg, distance=[10, 11] * u.kpc
    )
    # if the inputs are not the same, if possible they will be broadcast following
    # numpy's standard broadcasting rules.
    c2 = SphericalRepresentation(
        lon=[8, 9] * u.hourangle, lat=[5, 6] * u.deg, distance=10 * u.kpc
    )
    assert len(c2.distance) == 2
    # when they can't be broadcast, it is a ValueError (same as Numpy)
    with pytest.raises(ValueError):
        c2 = UnitSphericalRepresentation(
            lon=[8, 9, 10] * u.hourangle, lat=[5, 6] * u.deg
        )

    # It's also possible to pass in scalar quantity lists with mixed units. These
    # are converted to array quantities following the same rule as `Quantity`: all
    # elements are converted to match the first element's units.
    c2 = UnitSphericalRepresentation(
        lon=Angle([8 * u.hourangle, 135 * u.deg]),
        lat=Angle([5 * u.deg, (6 * np.pi / 180) * u.rad]),
    )
    assert c2.lat.unit == u.deg and c2.lon.unit == u.hourangle
    npt.assert_almost_equal(c2.lon[1].value, 9)

    # The Quantity initializer itself can also be used to force the unit even if the
    # first element doesn't have the right unit
    lon = u.Quantity([120 * u.deg, 135 * u.deg], u.hourangle)
    lat = u.Quantity([(5 * np.pi / 180) * u.rad, 0.4 * u.hourangle], u.deg)
    c2 = UnitSphericalRepresentation(lon, lat)

    # regardless of how input, the `lat` and `lon` come out as angle/distance
    assert isinstance(c1.lat, Angle)
    assert isinstance(
        c1.lat, Latitude
    )  # `Latitude` is an `~astropy.coordinates.Angle` subclass
    assert isinstance(c1.distance, Distance)

    # but they are read-only, as representations are immutable once created
    with pytest.raises(AttributeError):
        c1.lat = Latitude(5, u.deg)
    # Note that it is still possible to modify the array in-place, but this is not
    # sanctioned by the API, as this would prevent things like caching.
    c2.lat[:] = [0] * u.deg  # possible, but NOT SUPPORTED

    # To address the fact that there are various other conventions for how spherical
    # coordinates are defined, other conventions can be included as new classes.
    # Later there may be other conventions that we implement - for now just the
    # physics convention, as it is one of the most common cases.
    _ = PhysicsSphericalRepresentation(phi=120 * u.deg, theta=85 * u.deg, r=3 * u.kpc)

    # first dimension must be length-3 if a lone `Quantity` is passed in.
    c1 = CartesianRepresentation(np.random.randn(3, 100) * u.kpc)
    assert c1.xyz.shape[0] == 3
    assert c1.xyz.unit == u.kpc
    assert c1.x.shape[0] == 100
    assert c1.y.shape[0] == 100
    assert c1.z.shape[0] == 100
    # can also give each as separate keywords
    CartesianRepresentation(
        x=np.random.randn(100) * u.kpc,
        y=np.random.randn(100) * u.kpc,
        z=np.random.randn(100) * u.kpc,
    )
    # if the units don't match but are all distances, they will automatically be
    # converted to match `x`
    xarr, yarr, zarr = np.random.randn(3, 100)
    c1 = CartesianRepresentation(x=xarr * u.kpc, y=yarr * u.kpc, z=zarr * u.kpc)
    c2 = CartesianRepresentation(x=xarr * u.kpc, y=yarr * u.kpc, z=zarr * u.pc)
    assert c1.xyz.unit == c2.xyz.unit == u.kpc
    assert_allclose((c1.z / 1000) - c2.z, 0 * u.kpc, atol=1e-10 * u.kpc)

    # representations convert into other representations via  `represent_as`
    srep = SphericalRepresentation(lon=90 * u.deg, lat=0 * u.deg, distance=1 * u.pc)
    crep = srep.represent_as(CartesianRepresentation)
    assert_allclose(crep.x, 0 * u.pc, atol=1e-10 * u.pc)
    assert_allclose(crep.y, 1 * u.pc, atol=1e-10 * u.pc)
    assert_allclose(crep.z, 0 * u.pc, atol=1e-10 * u.pc)
    # The functions that actually do the conversion are defined via methods on the
    # representation classes. This may later be expanded into a full registerable
    # transform graph like the coordinate frames, but initially it will be a simpler
    # method system


def test_frame_api():
    from astropy.coordinates.builtin_frames import FK5, ICRS
    from astropy.coordinates.representation import (
        SphericalRepresentation,
        UnitSphericalRepresentation,
    )

    # <--------------------Reference Frame/"Low-level" classes--------------------->
    # The low-level classes have a dual role: they act as specifiers of coordinate
    # frames and they *may* also contain data as one of the representation objects,
    # in which case they are the actual coordinate objects themselves.
    # They can always accept a representation as a first argument
    icrs = ICRS(UnitSphericalRepresentation(lon=8 * u.hour, lat=5 * u.deg))

    # which is stored as the `data` attribute
    assert icrs.data.lat == 5 * u.deg
    assert icrs.data.lon == 8 * u.hourangle

    # Frames that require additional information like equinoxs or obstimes get them
    # as keyword parameters to the frame constructor.  Where sensible, defaults are
    # used. E.g., FK5 is almost always J2000 equinox
    fk5 = FK5(UnitSphericalRepresentation(lon=8 * u.hour, lat=5 * u.deg))
    J2000 = time.Time("J2000")
    fk5_2000 = FK5(
        UnitSphericalRepresentation(lon=8 * u.hour, lat=5 * u.deg), equinox=J2000
    )
    assert fk5.equinox == fk5_2000.equinox

    # the information required to specify the frame is immutable
    J2001 = time.Time("J2001")
    with pytest.raises(AttributeError):
        fk5.equinox = J2001

    # Similar for the representation data.
    with pytest.raises(AttributeError):
        fk5.data = UnitSphericalRepresentation(lon=8 * u.hour, lat=5 * u.deg)

    # There is also a class-level attribute that lists the attributes needed to
    # identify the frame.  These include attributes like `equinox` shown above.
    assert all(nm in ("equinox", "obstime") for nm in fk5.frame_attributes)

    # the `frame_attributes` are called for particularly in  the
    # high-level class (discussed below) to allow round-tripping between various
    # frames.  It is also part of the public API for other similar developer /
    # advanced users' use.

    # The actual position information is accessed via the representation objects
    assert_allclose(icrs.represent_as(SphericalRepresentation).lat, 5 * u.deg)
    # shorthand for the above
    assert_allclose(icrs.spherical.lat, 5 * u.deg)
    assert icrs.cartesian.z.value > 0

    # Many frames have a "default" representation, the one in which they are
    # conventionally described, often with a special name for some of the
    # coordinates. E.g., most equatorial coordinate systems are spherical with RA and
    # Dec. This works simply as a shorthand for the longer form above

    assert_allclose(icrs.dec, 5 * u.deg)
    assert_allclose(fk5.ra, 8 * u.hourangle)

    assert icrs.representation_type == SphericalRepresentation

    # low-level classes can also be initialized with names valid for that representation
    # and frame:
    icrs_2 = ICRS(ra=8 * u.hour, dec=5 * u.deg, distance=1 * u.kpc)
    assert_allclose(icrs.ra, icrs_2.ra)

    # and these are taken as the default if keywords are not given:
    # icrs_nokwarg = ICRS(8*u.hour, 5*u.deg, distance=1*u.kpc)
    # assert icrs_nokwarg.ra == icrs_2.ra and icrs_nokwarg.dec == icrs_2.dec

    # they also are capable of computing on-sky or 3d separations from each other,
    # which will be a direct port of the existing methods:
    coo1 = ICRS(ra=0 * u.hour, dec=0 * u.deg)
    coo2 = ICRS(ra=0 * u.hour, dec=1 * u.deg)
    # `separation` is the on-sky separation
    assert_allclose(coo1.separation(coo2).degree, 1.0)

    # while `separation_3d` includes the 3D distance information
    coo3 = ICRS(ra=0 * u.hour, dec=0 * u.deg, distance=1 * u.kpc)
    coo4 = ICRS(ra=0 * u.hour, dec=0 * u.deg, distance=2 * u.kpc)
    assert coo3.separation_3d(coo4).kpc == 1.0

    # The next example fails because `coo1` and `coo2` don't have distances
    with pytest.raises(ValueError):
        assert coo1.separation_3d(coo2).kpc == 1.0

    # repr/str also shows info, with frame and data
    # assert repr(fk5) == ''


def test_transform_api():
    from astropy.coordinates.baseframe import BaseCoordinateFrame, frame_transform_graph
    from astropy.coordinates.builtin_frames import FK5, ICRS
    from astropy.coordinates.representation import UnitSphericalRepresentation
    from astropy.coordinates.transformations import DynamicMatrixTransform

    # <------------------------Transformations------------------------------------->
    # Transformation functionality is the key to the whole scheme: they transform
    # low-level classes from one frame to another.
    # (used below but defined above in the API)
    fk5 = FK5(ra=8 * u.hour, dec=5 * u.deg)

    # If no data (or `None`) is given, the class acts as a specifier of a frame, but
    # without any stored data.
    J2001 = time.Time("J2001")
    fk5_J2001_frame = FK5(equinox=J2001)

    # if they do not have data, the string instead is the frame specification
    assert repr(fk5_J2001_frame) == "<FK5 Frame (equinox=J2001.000)>"

    #  Note that, although a frame object is immutable and can't have data added, it
    #  can be used to create a new object that does have data by giving the
    # `realize_frame` method a representation:
    srep = UnitSphericalRepresentation(lon=8 * u.hour, lat=5 * u.deg)
    fk5_j2001_with_data = fk5_J2001_frame.realize_frame(srep)
    assert fk5_j2001_with_data.data is not None
    # Now `fk5_j2001_with_data` is in the same frame as `fk5_J2001_frame`, but it
    # is an actual low-level coordinate, rather than a frame without data.

    # These frames are primarily useful for specifying what a coordinate should be
    # transformed *into*, as they are used by the `transform_to` method
    # E.g., this snippet precesses the point to the new equinox
    newfk5 = fk5.transform_to(fk5_J2001_frame)
    assert newfk5.equinox == J2001

    # transforming to a new frame necessarily loses framespec information if that
    # information is not applicable to the new frame.  This means transforms are not
    # always round-trippable:
    fk5_2 = FK5(ra=8 * u.hour, dec=5 * u.deg, equinox=J2001)
    ic_trans = fk5_2.transform_to(ICRS())

    # `ic_trans` does not have an `equinox`, so now when we transform back to FK5,
    # it's a *different* RA and Dec
    fk5_trans = ic_trans.transform_to(FK5())
    assert not allclose(fk5_2.ra, fk5_trans.ra, rtol=0, atol=1e-10 * u.deg)

    # But if you explicitly give the right equinox, all is fine
    fk5_trans_2 = fk5_2.transform_to(FK5(equinox=J2001))
    assert_allclose(fk5_2.ra, fk5_trans_2.ra, rtol=0, atol=1e-10 * u.deg)

    # Trying to transforming a frame with no data is of course an error:
    with pytest.raises(ValueError):
        FK5(equinox=J2001).transform_to(ICRS())

    # To actually define a new transformation, the same scheme as in the
    # 0.2/0.3 coordinates framework can be re-used - a graph of transform functions
    # connecting various coordinate classes together.  The main changes are:
    # 1) The transform functions now get the frame object they are transforming the
    #    current data into.
    # 2) Frames with additional information need to have a way to transform between
    #    objects of the same class, but with different framespecinfo values

    # An example transform function:
    class SomeNewSystem(BaseCoordinateFrame):
        pass

    @frame_transform_graph.transform(DynamicMatrixTransform, SomeNewSystem, FK5)
    def new_to_fk5(newobj, fk5frame):
        _ = newobj.obstime
        _ = fk5frame.equinox
        # ... build a *cartesian* transform matrix using `eq` that transforms from
        # the `newobj` frame as observed at `ot` to FK5 an equinox `eq`
        matrix = np.eye(3)
        return matrix

    # Other options for transform functions include one that simply returns the new
    # coordinate object, and one that returns a cartesian matrix but does *not*
    # require `newobj` or `fk5frame` - this allows optimization of the transform.


def test_highlevel_api():
    J2001 = time.Time("J2001")

    # <--------------------------"High-level" class-------------------------------->
    # The "high-level" class is intended to wrap the lower-level classes in such a
    # way that they can be round-tripped, as well as providing a variety of
    # convenience functionality.  This document is not intended to show *all* of the
    # possible high-level functionality, rather how the high-level classes are
    # initialized and interact with the low-level classes

    # this creates an object that contains an `ICRS` low-level class, initialized
    # identically to the first ICRS example further up.

    sc = coords.SkyCoord(
        coords.SphericalRepresentation(
            lon=8 * u.hour, lat=5 * u.deg, distance=1 * u.kpc
        ),
        frame="icrs",
    )

    # Other representations and `system` keywords delegate to the appropriate
    # low-level class. The already-existing registry for user-defined coordinates
    # will be used by `SkyCoordinate` to figure out what various the `system`
    # keyword actually means.

    sc = coords.SkyCoord(ra=8 * u.hour, dec=5 * u.deg, frame="icrs")
    sc = coords.SkyCoord(l=120 * u.deg, b=5 * u.deg, frame="galactic")

    # High-level classes can also be initialized directly from low-level objects
    sc = coords.SkyCoord(coords.ICRS(ra=8 * u.hour, dec=5 * u.deg))

    # The next example raises an error because the high-level class must always
    # have position data.
    with pytest.raises(ValueError):
        sc = coords.SkyCoord(coords.FK5(equinox=J2001))  # raises ValueError

    # similarly, the low-level object can always be accessed

    # this is how it's supposed to look, but sometimes the numbers get rounded in
    # funny ways
    # assert repr(sc.frame) == '<ICRS Coordinate: ra=120.0 deg, dec=5.0 deg>'
    rscf = repr(sc.frame)
    assert rscf.startswith("<ICRS Coordinate: (ra, dec) in deg")

    # and  the string representation will be inherited from the low-level class.

    # same deal, should loook like this, but different archituectures/ python
    # versions may round the numbers differently
    # assert repr(sc) == '<SkyCoord (ICRS): ra=120.0 deg, dec=5.0 deg>'
    rsc = repr(sc)
    assert rsc.startswith("<SkyCoord (ICRS): (ra, dec) in deg")

    # Supports a variety of possible complex string formats
    sc = coords.SkyCoord("8h00m00s +5d00m00.0s", frame="icrs")

    # In the next example, the unit is only needed b/c units are ambiguous.  In
    # general, we *never* accept ambiguity
    sc = coords.SkyCoord("8:00:00 +5:00:00.0", unit=(u.hour, u.deg), frame="icrs")

    # The next one would yield length-2 array coordinates, because of the comma

    sc = coords.SkyCoord(["8h 5d", "2°2′3″ 0.3rad"], frame="icrs")

    # It should also interpret common designation styles as a coordinate
    # NOT YET
    # sc = coords.SkyCoord('SDSS J123456.89-012345.6', frame='icrs')

    # but it should also be possible to provide formats for outputting to strings,
    # similar to `Time`.  This can be added right away or at a later date.

    # transformation is done the same as for low-level classes, which it delegates to

    sc_fk5_j2001 = sc.transform_to(coords.FK5(equinox=J2001))
    assert sc_fk5_j2001.equinox == J2001

    # The key difference is that the high-level class remembers frame information
    # necessary for round-tripping, unlike the low-level classes:
    sc1 = coords.SkyCoord(ra=8 * u.hour, dec=5 * u.deg, equinox=J2001, frame="fk5")
    sc2 = sc1.transform_to("icrs")

    # The next assertion succeeds, but it doesn't mean anything for ICRS, as ICRS
    # isn't defined in terms of an equinox
    assert sc2.equinox == J2001

    # But it *is* necessary once we transform to FK5
    sc3 = sc2.transform_to("fk5")
    assert sc3.equinox == J2001
    assert_allclose(sc1.ra, sc3.ra)

    # `SkyCoord` will also include the attribute-style access that is in the
    # v0.2/0.3 coordinate objects.  This will *not* be in the low-level classes
    sc = coords.SkyCoord(ra=8 * u.hour, dec=5 * u.deg, frame="icrs")
    scgal = sc.galactic
    assert str(scgal).startswith("<SkyCoord (Galactic): (l, b)")

    # the existing `from_name` and `match_to_catalog_*` methods will be moved to the
    # high-level class as convenience functionality.

    # in remote-data test below!
    # m31icrs = coords.SkyCoord.from_name('M31', frame='icrs')
    # assert str(m31icrs) == '<SkyCoord (ICRS) RA=10.68471 deg, Dec=41.26875 deg>'

    if HAS_SCIPY:
        cat1 = coords.SkyCoord(
            ra=[1, 2] * u.hr,
            dec=[3, 4.01] * u.deg,
            distance=[5, 6] * u.kpc,
            frame="icrs",
        )
        cat2 = coords.SkyCoord(
            ra=[1, 2, 2.01] * u.hr,
            dec=[3, 4, 5] * u.deg,
            distance=[5, 200, 6] * u.kpc,
            frame="icrs",
        )
        idx1, sep2d1, dist3d1 = cat1.match_to_catalog_sky(cat2)
        idx2, sep2d2, dist3d2 = cat1.match_to_catalog_3d(cat2)

        assert np.any(idx1 != idx2)

    # additional convenience functionality for the future should be added as methods
    # on `SkyCoord`, *not* the low-level classes.


@pytest.mark.remote_data
def test_highlevel_api_remote():
    m31icrs = coords.SkyCoord.from_name("M31", frame="icrs")

    m31str = str(m31icrs)
    assert m31str.startswith("<SkyCoord (ICRS): (ra, dec) in deg\n    (")
    assert m31str.endswith(")>")
    assert "10.68" in m31str
    assert "41.26" in m31str
    # The above is essentially a replacement of the below, but tweaked so that
    # small/moderate changes in what `from_name` returns don't cause the tests
    # to fail
    # assert str(m31icrs) == '<SkyCoord (ICRS): (ra, dec) in deg\n    (10.6847083, 41.26875)>'

    m31fk4 = coords.SkyCoord.from_name("M31", frame="fk4")

    assert not m31icrs.is_equivalent_frame(m31fk4)
    assert np.abs(m31icrs.ra - m31fk4.ra) > 0.5 * u.deg



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Classes to read AAS MRT table format.

Ref: https://journals.aas.org/mrt-standards

:Copyright: Smithsonian Astrophysical Observatory (2021)
:Author: Tom Aldcroft (aldcroft@head.cfa.harvard.edu), \
         Suyog Garg (suyog7130@gmail.com)
"""

import re
import warnings
from io import StringIO
from math import ceil, floor
from string import Template
from textwrap import wrap

import numpy as np

from astropy import units as u
from astropy.table import Column, MaskedColumn, Table

from . import cds, core, fixedwidth

MAX_SIZE_README_LINE = 80
MAX_COL_INTLIMIT = 100000


__doctest_skip__ = ["*"]


BYTE_BY_BYTE_TEMPLATE = [
    "Byte-by-byte Description of file: $file",
    "--------------------------------------------------------------------------------",
    " Bytes Format Units  Label     Explanations",
    "--------------------------------------------------------------------------------",
    "$bytebybyte",
    "--------------------------------------------------------------------------------",
]

MRT_TEMPLATE = [
    "Title:",
    "Authors:",
    "Table:",
    "================================================================================",
    "$bytebybyte",
    "Notes:",
    "--------------------------------------------------------------------------------",
]


class MrtSplitter(fixedwidth.FixedWidthSplitter):
    """
    Contains the join function to left align the MRT columns
    when writing to a file.
    """

    def join(self, vals, widths):
        vals = [val + " " * (width - len(val)) for val, width in zip(vals, widths)]
        return self.delimiter.join(vals)


class MrtHeader(cds.CdsHeader):
    _subfmt = "MRT"

    def _split_float_format(self, value):
        """
        Splits a Float string into different parts to find number
        of digits after decimal and check if the value is in Scientific
        notation.

        Parameters
        ----------
        value : str
            String containing the float value to split.

        Returns
        -------
        fmt: (int, int, int, bool, bool)
            List of values describing the Float string.
            (size, dec, ent, sign, exp)
            size, length of the given string.
            ent, number of digits before decimal point.
            dec, number of digits after decimal point.
            sign, whether or not given value signed.
            exp, is value in Scientific notation?
        """
        regfloat = re.compile(
            r"""(?P<sign> [+-]*)
                (?P<ent> [^eE.]+)
                (?P<deciPt> [.]*)
                (?P<decimals> [0-9]*)
                (?P<exp> [eE]*-*)[0-9]*""",
            re.VERBOSE,
        )
        mo = regfloat.match(value)

        if mo is None:
            raise Exception(f"{value} is not a float number")
        return (
            len(value),
            len(mo.group("ent")),
            len(mo.group("decimals")),
            mo.group("sign") != "",
            mo.group("exp") != "",
        )

    def _set_column_val_limits(self, col):
        """
        Sets the ``col.min`` and ``col.max`` column attributes,
        taking into account columns with Null values.
        """
        col.max = max(col)
        col.min = min(col)
        if col.max is np.ma.core.MaskedConstant:
            col.max = None
        if col.min is np.ma.core.MaskedConstant:
            col.min = None

    def column_float_formatter(self, col):
        """
        String formatter function for a column containing Float values.
        Checks if the values in the given column are in Scientific notation,
        by splitting the value string. It is assumed that the column either has
        float values or Scientific notation.

        A ``col.formatted_width`` attribute is added to the column. It is not added
        if such an attribute is already present, say when the ``formats`` argument
        is passed to the writer. A properly formatted format string is also added as
        the ``col.format`` attribute.

        Parameters
        ----------
        col : A ``Table.Column`` object.
        """
        # maxsize: maximum length of string containing the float value.
        # maxent: maximum number of digits places before decimal point.
        # maxdec: maximum number of digits places after decimal point.
        # maxprec: maximum precision of the column values, sum of maxent and maxdec.
        maxsize, maxprec, maxent, maxdec = 1, 0, 1, 0
        sign = False
        fformat = "F"

        # Find maximum sized value in the col
        for val in col.str_vals:
            # Skip null values
            if val is None or val == "":
                continue

            # Find format of the Float string
            fmt = self._split_float_format(val)
            # If value is in Scientific notation
            if fmt[4] is True:
                # if the previous column value was in normal Float format
                # set maxsize, maxprec and maxdec to default.
                if fformat == "F":
                    maxsize, maxprec, maxdec = 1, 0, 0
                # Designate the column to be in Scientific notation.
                fformat = "E"
            else:
                # Move to next column value if
                # current value is not in Scientific notation
                # but the column is designated as such because
                # one of the previous values was.
                if fformat == "E":
                    continue

            if maxsize < fmt[0]:
                maxsize = fmt[0]
            if maxent < fmt[1]:
                maxent = fmt[1]
            if maxdec < fmt[2]:
                maxdec = fmt[2]
            if fmt[3]:
                sign = True

            if maxprec < fmt[1] + fmt[2]:
                maxprec = fmt[1] + fmt[2]

        if fformat == "E":
            # If ``formats`` not passed.
            if getattr(col, "formatted_width", None) is None:
                col.formatted_width = maxsize
                if sign:
                    col.formatted_width += 1
            # Number of digits after decimal is replaced by the precision
            # for values in Scientific notation, when writing that Format.
            col.fortran_format = fformat + str(col.formatted_width) + "." + str(maxprec)
            col.format = str(col.formatted_width) + "." + str(maxdec) + "e"
        else:
            lead = ""
            if (
                getattr(col, "formatted_width", None) is None
            ):  # If ``formats`` not passed.
                col.formatted_width = maxent + maxdec + 1
                if sign:
                    col.formatted_width += 1
            elif col.format.startswith("0"):
                # Keep leading zero, if already set in format - primarily for `seconds` columns
                # in coordinates; may need extra case if this is to be also supported with `sign`.
                lead = "0"
            col.fortran_format = fformat + str(col.formatted_width) + "." + str(maxdec)
            col.format = lead + col.fortran_format[1:] + "f"

    def write_byte_by_byte(self):
        """
        Writes the Byte-By-Byte description of the table.

        Columns that are `astropy.coordinates.SkyCoord` or `astropy.time.TimeSeries`
        objects or columns with values that are such objects are recognized as such,
        and some predefined labels and description is used for them.
        See the Vizier MRT Standard documentation in the link below for more details
        on these. An example Byte-By-Byte table is shown here.

        See: https://vizier.unistra.fr/doc/catstd-3.1.htx

        Example::

        --------------------------------------------------------------------------------
        Byte-by-byte Description of file: table.dat
        --------------------------------------------------------------------------------
        Bytes Format Units  Label     Explanations
        --------------------------------------------------------------------------------
         1- 8  A8     ---    names   Description of names
        10-14  E5.1   ---    e       [-3160000.0/0.01] Description of e
        16-23  F8.5   ---    d       [22.25/27.25] Description of d
        25-31  E7.1   ---    s       [-9e+34/2.0] Description of s
        33-35  I3     ---    i       [-30/67] Description of i
        37-39  F3.1   ---    sameF   [5.0/5.0] Description of sameF
        41-42  I2     ---    sameI   [20] Description of sameI
        44-45  I2     h      RAh     Right Ascension (hour)
        47-48  I2     min    RAm     Right Ascension (minute)
        50-67  F18.15 s      RAs     Right Ascension (second)
           69  A1     ---    DE-     Sign of Declination
        70-71  I2     deg    DEd     Declination (degree)
        73-74  I2     arcmin DEm     Declination (arcmin)
        76-91  F16.13 arcsec DEs     Declination (arcsec)

        --------------------------------------------------------------------------------
        """
        # Get column widths
        vals_list = list(zip(*self.data.str_vals()))

        for i, col in enumerate(self.cols):
            col.width = max(len(vals[i]) for vals in vals_list)
            if self.start_line is not None:
                col.width = max(col.width, len(col.info.name))
        widths = [col.width for col in self.cols]

        startb = 1  # Byte count starts at 1.

        # Set default width of the Bytes count column of the Byte-By-Byte table.
        # This ``byte_count_width`` value helps align byte counts with respect
        # to the hyphen using a format string.
        byte_count_width = len(str(sum(widths) + len(self.cols) - 1))

        # Format string for Start Byte and End Byte
        singlebfmt = "{:" + str(byte_count_width) + "d}"
        fmtb = singlebfmt + "-" + singlebfmt
        # Add trailing single whitespaces to Bytes column for better visibility.
        singlebfmt += " "
        fmtb += " "

        # Set default width of Label and Description Byte-By-Byte columns.
        max_label_width, max_descrip_size = 7, 16

        bbb = Table(
            names=["Bytes", "Format", "Units", "Label", "Explanations"], dtype=[str] * 5
        )

        # Iterate over the columns to write Byte-By-Byte rows.
        for i, col in enumerate(self.cols):
            # Check if column is MaskedColumn
            col.has_null = isinstance(col, MaskedColumn)

            if col.format is not None:
                col.formatted_width = max(len(sval) for sval in col.str_vals)

            # Set MRTColumn type, size and format.
            if np.issubdtype(col.dtype, np.integer):
                # Integer formatter
                self._set_column_val_limits(col)
                # If ``formats`` not passed.
                if getattr(col, "formatted_width", None) is None:
                    col.formatted_width = max(len(str(col.max)), len(str(col.min)))
                col.fortran_format = "I" + str(col.formatted_width)
                if col.format is None:
                    col.format = ">" + col.fortran_format[1:]

            elif np.issubdtype(col.dtype, np.dtype(float).type):
                # Float formatter
                self._set_column_val_limits(col)
                self.column_float_formatter(col)

            else:
                # String formatter, ``np.issubdtype(col.dtype, str)`` is ``True``.
                dtype = col.dtype.str
                if col.has_null:
                    mcol = col
                    mcol.fill_value = ""
                    coltmp = Column(mcol.filled(), dtype=str)
                    dtype = coltmp.dtype.str
                # If ``formats`` not passed.
                if getattr(col, "formatted_width", None) is None:
                    col.formatted_width = int(re.search(r"(\d+)$", dtype).group(1))
                col.fortran_format = "A" + str(col.formatted_width)
                col.format = str(col.formatted_width) + "s"

            endb = col.formatted_width + startb - 1

            # ``mixin`` columns converted to string valued columns will not have a name
            # attribute. In those cases, a ``Unknown`` column label is put, indicating that
            # such columns can be better formatted with some manipulation before calling
            # the MRT writer.
            if col.name is None:
                col.name = "Unknown"

            # Set column description.
            if col.description is not None:
                description = col.description
            else:
                description = "Description of " + col.name

            # Set null flag in column description
            nullflag = ""
            if col.has_null:
                nullflag = "?"

            # Set column unit
            if col.unit is not None:
                col_unit = col.unit.to_string("cds")
            elif col.name.lower().find("magnitude") > -1:
                # ``col.unit`` can still be ``None``, if the unit of column values
                # is ``Magnitude``, because ``astropy.units.Magnitude`` is actually a class.
                # Unlike other units which are instances of ``astropy.units.Unit``,
                # application of the ``Magnitude`` unit calculates the logarithm
                # of the values. Thus, the only way to check for if the column values
                # have ``Magnitude`` unit is to check the column name.
                col_unit = "mag"
            else:
                col_unit = "---"

            # Add col limit values to col description
            lim_vals = ""
            if (
                col.min
                and col.max
                and not any(
                    x in col.name for x in ["RA", "DE", "LON", "LAT", "PLN", "PLT"]
                )
            ):
                # No col limit values for coordinate columns.
                if col.fortran_format[0] == "I":
                    if (
                        abs(col.min) < MAX_COL_INTLIMIT
                        and abs(col.max) < MAX_COL_INTLIMIT
                    ):
                        if col.min == col.max:
                            lim_vals = f"[{col.min}]"
                        else:
                            lim_vals = f"[{col.min}/{col.max}]"
                elif col.fortran_format[0] in ("E", "F"):
                    lim_vals = (
                        f"[{floor(col.min * 100) / 100.}/{ceil(col.max * 100) / 100.}]"
                    )

            if lim_vals != "" or nullflag != "":
                description = f"{lim_vals}{nullflag} {description}"

            # Find the maximum label and description column widths.
            if len(col.name) > max_label_width:
                max_label_width = len(col.name)
            if len(description) > max_descrip_size:
                max_descrip_size = len(description)

            # Add a row for the Sign of Declination in the bbb table
            if col.name == "DEd":
                bbb.add_row(
                    [
                        singlebfmt.format(startb),
                        "A1",
                        "---",
                        "DE-",
                        "Sign of Declination",
                    ]
                )
                col.fortran_format = "I2"
                startb += 1

            # Add Byte-By-Byte row to bbb table
            bbb.add_row(
                [
                    singlebfmt.format(startb)
                    if startb == endb
                    else fmtb.format(startb, endb),
                    "" if col.fortran_format is None else col.fortran_format,
                    col_unit,
                    "" if col.name is None else col.name,
                    description,
                ]
            )
            startb = endb + 2

        # Properly format bbb columns
        bbblines = StringIO()
        bbb.write(
            bbblines,
            format="ascii.fixed_width_no_header",
            delimiter=" ",
            bookend=False,
            delimiter_pad=None,
            formats={
                "Format": "<6s",
                "Units": "<6s",
                "Label": "<" + str(max_label_width) + "s",
                "Explanations": "" + str(max_descrip_size) + "s",
            },
        )

        # Get formatted bbb lines
        bbblines = bbblines.getvalue().splitlines()

        # ``nsplit`` is the number of whitespaces to prefix to long description
        # lines in order to wrap them. It is the sum of the widths of the
        # previous 4 columns plus the number of single spacing between them.
        # The hyphen in the Bytes column is also counted.
        nsplit = byte_count_width * 2 + 1 + 12 + max_label_width + 4

        # Wrap line if it is too long
        buff = ""
        for newline in bbblines:
            if len(newline) > MAX_SIZE_README_LINE:
                buff += ("\n").join(
                    wrap(
                        newline,
                        subsequent_indent=" " * nsplit,
                        width=MAX_SIZE_README_LINE,
                    )
                )
                buff += "\n"
            else:
                buff += newline + "\n"

        # Last value of ``endb`` is the sum of column widths after formatting.
        self.linewidth = endb

        # Remove the last extra newline character from Byte-By-Byte.
        buff = buff[:-1]
        return buff

    def write(self, lines):
        """
        Writes the Header of the MRT table, aka ReadMe, which
        also contains the Byte-By-Byte description of the table.
        """
        from astropy.coordinates import SkyCoord

        # Recognised ``SkyCoord.name`` forms with their default column names (helio* require SunPy).
        coord_systems = {
            "galactic": ("GLAT", "GLON", "b", "l"),
            "ecliptic": ("ELAT", "ELON", "lat", "lon"),  # 'geocentric*ecliptic'
            "heliographic": ("HLAT", "HLON", "lat", "lon"),  # '_carrington|stonyhurst'
            "helioprojective": ("HPLT", "HPLN", "Ty", "Tx"),
        }
        eqtnames = ["RAh", "RAm", "RAs", "DEd", "DEm", "DEs"]

        # list to store indices of columns that are modified.
        to_pop = []

        # For columns that are instances of ``SkyCoord`` and other ``mixin`` columns
        # or whose values are objects of these classes.
        for i, col in enumerate(self.cols):
            # If col is a ``Column`` object but its values are ``SkyCoord`` objects,
            # convert the whole column to ``SkyCoord`` object, which helps in applying
            # SkyCoord methods directly.
            if not isinstance(col, SkyCoord) and isinstance(col[0], SkyCoord):
                try:
                    col = SkyCoord(col)
                except (ValueError, TypeError):
                    # If only the first value of the column is a ``SkyCoord`` object,
                    # the column cannot be converted to a ``SkyCoord`` object.
                    # These columns are converted to ``Column`` object and then converted
                    # to string valued column.
                    if not isinstance(col, Column):
                        col = Column(col)
                    col = Column([str(val) for val in col])
                    self.cols[i] = col
                    continue

            # Replace single ``SkyCoord`` column by its coordinate components if no coordinate
            # columns of the corresponding type exist yet.
            if isinstance(col, SkyCoord):
                # If coordinates are given in RA/DEC, divide each them into hour/deg,
                # minute/arcminute, second/arcsecond columns.
                if (
                    "ra" in col.representation_component_names.keys()
                    and len(set(eqtnames) - set(self.colnames)) == 6
                ):
                    ra_c, dec_c = col.ra.hms, col.dec.dms
                    coords = [
                        ra_c.h.round().astype("i1"),
                        ra_c.m.round().astype("i1"),
                        ra_c.s,
                        dec_c.d.round().astype("i1"),
                        dec_c.m.round().astype("i1"),
                        dec_c.s,
                    ]
                    coord_units = [u.h, u.min, u.second, u.deg, u.arcmin, u.arcsec]
                    coord_descrip = [
                        "Right Ascension (hour)",
                        "Right Ascension (minute)",
                        "Right Ascension (second)",
                        "Declination (degree)",
                        "Declination (arcmin)",
                        "Declination (arcsec)",
                    ]
                    for coord, name, coord_unit, descrip in zip(
                        coords, eqtnames, coord_units, coord_descrip
                    ):
                        # Have Sign of Declination only in the DEd column.
                        if name in ["DEm", "DEs"]:
                            coord_col = Column(
                                list(np.abs(coord)),
                                name=name,
                                unit=coord_unit,
                                description=descrip,
                            )
                        else:
                            coord_col = Column(
                                list(coord),
                                name=name,
                                unit=coord_unit,
                                description=descrip,
                            )
                        # Set default number of digits after decimal point for the
                        # second values, and deg-min to (signed) 2-digit zero-padded integer.
                        if name == "RAs":
                            coord_col.format = "013.10f"
                        elif name == "DEs":
                            coord_col.format = "012.9f"
                        elif name == "RAh":
                            coord_col.format = "2d"
                        elif name == "DEd":
                            coord_col.format = "+03d"
                        elif name.startswith(("RA", "DE")):
                            coord_col.format = "02d"
                        self.cols.append(coord_col)
                    to_pop.append(i)  # Delete original ``SkyCoord`` column.

                # For all other coordinate types, simply divide into two columns
                # for latitude and longitude resp. with the unit used been as it is.

                else:
                    frminfo = ""
                    for frame, latlon in coord_systems.items():
                        if (
                            frame in col.name
                            and len(set(latlon[:2]) - set(self.colnames)) == 2
                        ):
                            if frame != col.name:
                                frminfo = f" ({col.name})"
                            lon_col = Column(
                                getattr(col, latlon[3]),
                                name=latlon[1],
                                description=f"{frame.capitalize()} Longitude{frminfo}",
                                unit=col.representation_component_units[latlon[3]],
                                format=".12f",
                            )
                            lat_col = Column(
                                getattr(col, latlon[2]),
                                name=latlon[0],
                                description=f"{frame.capitalize()} Latitude{frminfo}",
                                unit=col.representation_component_units[latlon[2]],
                                format="+.12f",
                            )
                            self.cols.append(lon_col)
                            self.cols.append(lat_col)
                            to_pop.append(i)  # Delete original ``SkyCoord`` column.

                # Convert all other ``SkyCoord`` columns that are not in the above three
                # representations to string valued columns. Those could either be types not
                # supported yet (e.g. 'helioprojective'), or already present and converted.
                # If there were any extra ``SkyCoord`` columns of one kind after the first one,
                # then their decomposition into their component columns has been skipped.
                # This is done in order to not create duplicate component columns.
                # Explicit renaming of the extra coordinate component columns by appending some
                # suffix to their name, so as to distinguish them, is not yet implemented.
                if i not in to_pop:
                    warnings.warn(
                        f"Coordinate system of type '{col.name}' already stored in"
                        " table as CDS/MRT-syle columns or of unrecognized type. So"
                        f" column {i} is being skipped with designation of a string"
                        f" valued column `{self.colnames[i]}`.",
                        UserWarning,
                    )
                    self.cols.append(Column(col.to_string(), name=self.colnames[i]))
                    to_pop.append(i)  # Delete original ``SkyCoord`` column.

            # Convert all other ``mixin`` columns to ``Column`` objects.
            # Parsing these may still lead to errors!
            elif not isinstance(col, Column):
                col = Column(col, name=self.colnames[i])
                # If column values are ``object`` types, convert them to string.
                if np.issubdtype(col.dtype, np.dtype(object).type):
                    col = Column([str(val) for val in col], name=col.name)
                self.cols[i] = col

        # Delete original ``SkyCoord`` columns, if there were any.
        for i in to_pop[::-1]:
            self.cols.pop(i)

        # Check for any left over extra coordinate columns.
        if any(x in self.colnames for x in ["RAh", "DEd", "ELON", "GLAT"]):
            # At this point any extra ``SkyCoord`` columns should have been converted to string
            # valued columns, together with issuance of a warning, by the coordinate parser above.
            # This test is just left here as a safeguard.
            for i, col in enumerate(self.cols):
                if isinstance(col, SkyCoord):
                    self.cols[i] = Column(col.to_string(), name=self.colnames[i])
                    message = (
                        "Table already has coordinate system in CDS/MRT-syle columns. "
                        f"So column {i} should have been replaced already with "
                        f"a string valued column `{self.colnames[i]}`."
                    )
                    raise core.InconsistentTableError(message)

        # Get Byte-By-Byte description and fill the template
        bbb_template = Template("\n".join(BYTE_BY_BYTE_TEMPLATE))
        byte_by_byte = bbb_template.substitute(
            {"file": "table.dat", "bytebybyte": self.write_byte_by_byte()}
        )

        # Fill up the full ReadMe
        rm_template = Template("\n".join(MRT_TEMPLATE))
        readme_filled = rm_template.substitute({"bytebybyte": byte_by_byte})
        lines.append(readme_filled)


class MrtData(cds.CdsData):
    """MRT table data reader."""

    _subfmt = "MRT"
    splitter_class = MrtSplitter

    def write(self, lines):
        self.splitter.delimiter = " "
        fixedwidth.FixedWidthData.write(self, lines)


class Mrt(core.BaseReader):
    """AAS MRT (Machine-Readable Table) format table.

    **Reading**
    ::

      >>> from astropy.io import ascii
      >>> table = ascii.read('data.mrt', format='mrt')

    **Writing**

    Use ``ascii.write(table, 'data.mrt', format='mrt')`` to  write tables to
    Machine Readable Table (MRT) format.

    Note that the metadata of the table, apart from units, column names and
    description, will not be written. These have to be filled in by hand later.

    See also: :ref:`cds_mrt_format`.

    Caveats:

    * The Units and Explanations are available in the column ``unit`` and
      ``description`` attributes, respectively.
    * The other metadata defined by this format is not available in the output table.
    """

    _format_name = "mrt"
    _io_registry_format_aliases = ["mrt"]
    _io_registry_can_write = True
    _description = "MRT format table"

    data_class = MrtData
    header_class = MrtHeader

    def write(self, table=None):
        # Construct for writing empty table is not yet done.
        if len(table) == 0:
            raise NotImplementedError

        self.data.header = self.header
        self.header.position_line = None
        self.header.start_line = None

        # Create a copy of the ``table``, so that it the copy gets modified and
        # written to the file, while the original table remains as it is.
        table = table.copy()
        return super().write(table)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


import warnings
from collections import OrderedDict
from copy import deepcopy
from io import StringIO

import numpy as np
import pytest

from astropy import coordinates, table, time
from astropy import units as u
from astropy.table.info import serialize_method_as
from astropy.table.table_helpers import simple_table
from astropy.utils.data_info import data_info_factory, dtype_info_name


def test_table_info_attributes(table_types):
    """
    Test the info() method of printing a summary of table column attributes
    """
    a = np.array([1, 2, 3], dtype="int32")
    b = np.array([1, 2, 3], dtype="float32")
    c = np.array(["a", "c", "e"], dtype="|S1")
    t = table_types.Table([a, b, c], names=["a", "b", "c"])

    # Minimal output for a typical table
    tinfo = t.info(out=None)
    subcls = ["class"] if table_types.Table.__name__ == "MyTable" else []
    assert tinfo.colnames == [
        "name",
        "dtype",
        "shape",
        "unit",
        "format",
        "description",
        "class",
        "n_bad",
        "length",
    ]
    assert np.all(tinfo["name"] == ["a", "b", "c"])
    assert np.all(tinfo["dtype"] == ["int32", "float32", dtype_info_name("S1")])
    if subcls:
        assert np.all(tinfo["class"] == ["MyColumn"] * 3)

    # All output fields including a mixin column
    t["d"] = [1, 2, 3] * u.m
    t["d"].description = "quantity"
    t["a"].format = "%02d"
    t["e"] = time.Time([1, 2, 3], format="mjd")
    t["e"].info.description = "time"
    t["f"] = coordinates.SkyCoord([1, 2, 3], [1, 2, 3], unit="deg")
    t["f"].info.description = "skycoord"

    tinfo = t.info(out=None)
    assert np.all(tinfo["name"] == "a b c d e f".split())
    assert np.all(
        tinfo["dtype"]
        == ["int32", "float32", dtype_info_name("S1"), "float64", "object", "object"]
    )
    assert np.all(tinfo["unit"] == ["", "", "", "m", "", "deg,deg"])
    assert np.all(tinfo["format"] == ["%02d", "", "", "", "", ""])
    assert np.all(tinfo["description"] == ["", "", "", "quantity", "time", "skycoord"])
    cls = t.ColumnClass.__name__
    assert np.all(tinfo["class"] == [cls, cls, cls, cls, "Time", "SkyCoord"])

    # Test that repr(t.info) is same as t.info()
    out = StringIO()
    t.info(out=out)
    assert repr(t.info) == out.getvalue()


def test_table_info_stats(table_types):
    """
    Test the info() method of printing a summary of table column statistics
    """
    a = np.array([1, 2, 1, 2], dtype="int32")
    b = np.array([1, 2, 1, 2], dtype="float32")
    c = np.array(["a", "c", "e", "f"], dtype="|S1")
    d = time.Time([1, 2, 1, 2], format="mjd", scale="tai")
    t = table_types.Table([a, b, c, d], names=["a", "b", "c", "d"])

    # option = 'stats'
    masked = "masked=True " if t.masked else ""
    out = StringIO()
    t.info("stats", out=out)
    table_header_line = f"<{t.__class__.__name__} {masked}length=4>"
    exp = [
        table_header_line,
        "name mean std min max",
        "---- ---- --- --- ---",
        "   a  1.5 0.5   1   2",
        "   b  1.5 0.5   1   2",
        "   c   --  --  --  --",
        "   d  1.5  -- 1.0 2.0",
    ]
    assert out.getvalue().splitlines() == exp

    # option = ['attributes', 'stats']
    tinfo = t.info(["attributes", "stats"], out=None)
    assert tinfo.colnames == [
        "name",
        "dtype",
        "shape",
        "unit",
        "format",
        "description",
        "class",
        "mean",
        "std",
        "min",
        "max",
        "n_bad",
        "length",
    ]
    assert np.all(tinfo["mean"] == ["1.5", "1.5", "--", "1.5"])
    assert np.all(tinfo["std"] == ["0.5", "0.5", "--", "--"])
    assert np.all(tinfo["min"] == ["1", "1", "--", "1.0"])
    assert np.all(tinfo["max"] == ["2", "2", "--", "2.0"])

    out = StringIO()
    t.info("stats", out=out)
    exp = [
        table_header_line,
        "name mean std min max",
        "---- ---- --- --- ---",
        "   a  1.5 0.5   1   2",
        "   b  1.5 0.5   1   2",
        "   c   --  --  --  --",
        "   d  1.5  -- 1.0 2.0",
    ]
    assert out.getvalue().splitlines() == exp

    # option = ['attributes', custom]
    custom = data_info_factory(
        names=["sum", "first"], funcs=[np.sum, lambda col: col[0]]
    )
    out = StringIO()
    tinfo = t.info(["attributes", custom], out=None)
    assert tinfo.colnames == [
        "name",
        "dtype",
        "shape",
        "unit",
        "format",
        "description",
        "class",
        "sum",
        "first",
        "n_bad",
        "length",
    ]
    assert np.all(tinfo["name"] == ["a", "b", "c", "d"])
    assert np.all(
        tinfo["dtype"] == ["int32", "float32", dtype_info_name("S1"), "object"]
    )
    assert np.all(tinfo["sum"] == ["6", "6", "--", "--"])
    assert np.all(tinfo["first"] == ["1", "1", "a", "1.0"])


def test_data_info():
    """
    Test getting info for just a column.
    """
    cols = [
        table.Column(
            [1.0, 2.0, np.nan], name="name", description="description", unit="m/s"
        ),
        table.MaskedColumn(
            [1.0, 2.0, 3.0],
            name="name",
            description="description",
            unit="m/s",
            mask=[False, False, True],
        ),
    ]
    for c in cols:
        # Test getting the full ordered dict
        cinfo = c.info(out=None)
        assert cinfo == OrderedDict(
            [
                ("name", "name"),
                ("dtype", "float64"),
                ("shape", ""),
                ("unit", "m / s"),
                ("format", ""),
                ("description", "description"),
                ("class", type(c).__name__),
                ("n_bad", 1),
                ("length", 3),
            ]
        )

        # Test the console (string) version which omits trivial values
        out = StringIO()
        c.info(out=out)
        exp = [
            "name = name",
            "dtype = float64",
            "unit = m / s",
            "description = description",
            f"class = {type(c).__name__}",
            "n_bad = 1",
            "length = 3",
        ]
        assert out.getvalue().splitlines() == exp

        # repr(c.info) gives the same as c.info()
        assert repr(c.info) == out.getvalue()

        # Test stats info
        cinfo = c.info("stats", out=None)
        assert cinfo == OrderedDict(
            [
                ("name", "name"),
                ("mean", "1.5"),
                ("std", "0.5"),
                ("min", "1"),
                ("max", "2"),
                ("n_bad", 1),
                ("length", 3),
            ]
        )


def test_data_info_subclass():
    class Column(table.Column):
        """
        Confusingly named Column on purpose, but that is legal.
        """

    for data in ([], [1, 2]):
        c = Column(data, dtype="int64")
        cinfo = c.info(out=None)
        assert cinfo == OrderedDict(
            [
                ("dtype", "int64"),
                ("shape", ""),
                ("unit", ""),
                ("format", ""),
                ("description", ""),
                ("class", "Column"),
                ("n_bad", 0),
                ("length", len(data)),
            ]
        )


def test_scalar_info():
    """
    Make sure info works with scalar values
    """
    c = time.Time("2000:001")
    cinfo = c.info(out=None)
    assert cinfo["n_bad"] == 0
    assert "length" not in cinfo


def test_empty_table():
    t = table.Table()
    out = StringIO()
    t.info(out=out)
    exp = ["<Table length=0>", "<No columns>"]
    assert out.getvalue().splitlines() == exp


def test_class_attribute():
    """
    Test that class info column is suppressed only for identical non-mixin
    columns.
    """
    vals = [[1] * u.m, [2] * u.m]

    texp = [
        "<Table length=1>",
        "name  dtype  unit",
        "---- ------- ----",
        "col0 float64    m",
        "col1 float64    m",
    ]

    qexp = [
        "<QTable length=1>",
        "name  dtype  unit  class  ",
        "---- ------- ---- --------",
        "col0 float64    m Quantity",
        "col1 float64    m Quantity",
    ]

    for table_cls, exp in ((table.Table, texp), (table.QTable, qexp)):
        t = table_cls(vals)
        out = StringIO()
        t.info(out=out)
        assert out.getvalue().splitlines() == exp


def test_ignore_warnings():
    t = table.Table([[np.nan, np.nan]])
    with warnings.catch_warnings(record=True) as warns:
        t.info("stats", out=None)
        assert len(warns) == 0


def test_no_deprecation_warning():
    # regression test for #5459, where numpy deprecation warnings were
    # emitted unnecessarily.
    t = simple_table()
    with warnings.catch_warnings(record=True) as warns:
        t.info()
        assert len(warns) == 0


def test_lost_parent_error():
    c = table.Column([1, 2, 3], name="a")
    with pytest.raises(AttributeError, match='failed to access "info" attribute'):
        c[:].info.name


def test_info_serialize_method():
    """
    Unit test of context manager to set info.serialize_method.  Normally just
    used to set this for writing a Table to file (FITS, ECSV, HDF5).
    """
    t = table.Table(
        {
            "tm": time.Time([1, 2], format="cxcsec"),
            "sc": coordinates.SkyCoord([1, 2], [1, 2], unit="deg"),
            "mc": table.MaskedColumn([1, 2], mask=[True, False]),
            "mc2": table.MaskedColumn([1, 2], mask=[True, False]),
        }
    )

    origs = {}
    for name in ("tm", "mc", "mc2"):
        origs[name] = deepcopy(t[name].info.serialize_method)

    # Test setting by name and getting back to originals
    with serialize_method_as(t, {"tm": "test_tm", "mc": "test_mc"}):
        for name in ("tm", "mc"):
            assert all(
                t[name].info.serialize_method[key] == "test_" + name
                for key in t[name].info.serialize_method
            )
        assert t["mc2"].info.serialize_method == origs["mc2"]
        assert not hasattr(t["sc"].info, "serialize_method")

    for name in ("tm", "mc", "mc2"):
        assert t[name].info.serialize_method == origs[name]  # dict compare
    assert not hasattr(t["sc"].info, "serialize_method")

    # Test setting by name and class, where name takes precedence.  Also
    # test that it works for subclasses.
    with serialize_method_as(
        t, {"tm": "test_tm", "mc": "test_mc", table.Column: "test_mc2"}
    ):
        for name in ("tm", "mc", "mc2"):
            assert all(
                t[name].info.serialize_method[key] == "test_" + name
                for key in t[name].info.serialize_method
            )
        assert not hasattr(t["sc"].info, "serialize_method")

    for name in ("tm", "mc", "mc2"):
        assert t[name].info.serialize_method == origs[name]  # dict compare
    assert not hasattr(t["sc"].info, "serialize_method")

    # Test supplying a single string that all applies to all columns with
    # a serialize_method.
    with serialize_method_as(t, "test"):
        for name in ("tm", "mc", "mc2"):
            assert all(
                t[name].info.serialize_method[key] == "test"
                for key in t[name].info.serialize_method
            )
        assert not hasattr(t["sc"].info, "serialize_method")

    for name in ("tm", "mc", "mc2"):
        assert t[name].info.serialize_method == origs[name]  # dict compare
    assert not hasattr(t["sc"].info, "serialize_method")


def test_info_serialize_method_exception():
    """
    Unit test of context manager to set info.serialize_method.  Normally just
    used to set this for writing a Table to file (FITS, ECSV, HDF5).
    """
    t = simple_table(masked=True)
    origs = deepcopy(t["a"].info.serialize_method)
    try:
        with serialize_method_as(t, "test"):
            assert all(
                t["a"].info.serialize_method[key] == "test"
                for key in t["a"].info.serialize_method
            )
            raise ZeroDivisionError()
    except ZeroDivisionError:
        pass

    assert t["a"].info.serialize_method == origs  # dict compare



# Licensed under a 3-clause BSD style license - see PYFITS.rst

import math
import os
import re
import time
from io import BytesIO
from itertools import product

import numpy as np
import pytest
from hypothesis import given
from hypothesis.extra.numpy import basic_indices
from numpy.testing import assert_equal

from astropy.io import fits
from astropy.io.fits.hdu.compressed import (
    COMPRESSION_TYPES,
    DITHER_SEED_CHECKSUM,
    SUBTRACTIVE_DITHER_1,
)
from astropy.io.fits.tests.conftest import FitsTestCase
from astropy.io.fits.tests.test_table import comparerecords
from astropy.utils.data import download_file
from astropy.utils.exceptions import AstropyDeprecationWarning


class TestCompressedImage(FitsTestCase):
    def test_empty(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2595
        """

        hdu = fits.CompImageHDU()
        assert hdu.data is None
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits"), mode="update") as hdul:
            assert len(hdul) == 2
            assert isinstance(hdul[1], fits.CompImageHDU)
            assert hdul[1].data is None

            # Now test replacing the empty data with an array and see what
            # happens
            hdul[1].data = np.arange(100, dtype=np.int32)

        with fits.open(self.temp("test.fits")) as hdul:
            assert len(hdul) == 2
            assert isinstance(hdul[1], fits.CompImageHDU)
            assert np.all(hdul[1].data == np.arange(100, dtype=np.int32))

    @pytest.mark.parametrize(
        ("data", "compression_type", "quantize_level"),
        [
            (np.zeros((2, 10, 10), dtype=np.float32), "RICE_1", 16),
            (np.zeros((2, 10, 10), dtype=np.float32), "GZIP_1", -0.01),
            (np.zeros((2, 10, 10), dtype=np.float32), "GZIP_2", -0.01),
            (np.zeros((100, 100)) + 1, "HCOMPRESS_1", 16),
            (np.zeros((10, 10)), "PLIO_1", 16),
        ],
    )
    @pytest.mark.parametrize("byte_order", ["<", ">"])
    def test_comp_image(self, data, compression_type, quantize_level, byte_order):
        data = data.view(data.dtype.newbyteorder(byte_order))
        primary_hdu = fits.PrimaryHDU()
        ofd = fits.HDUList(primary_hdu)
        chdu = fits.CompImageHDU(
            data,
            name="SCI",
            compression_type=compression_type,
            quantize_level=quantize_level,
        )
        ofd.append(chdu)
        ofd.writeto(self.temp("test_new.fits"), overwrite=True)
        ofd.close()
        with fits.open(self.temp("test_new.fits")) as fd:
            assert (fd[1].data == data).all()
            assert fd[1].header["NAXIS"] == chdu.header["NAXIS"]
            assert fd[1].header["NAXIS1"] == chdu.header["NAXIS1"]
            assert fd[1].header["NAXIS2"] == chdu.header["NAXIS2"]
            assert fd[1].header["BITPIX"] == chdu.header["BITPIX"]

    @pytest.mark.remote_data
    def test_comp_image_quantize_level(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/5969

        Test that quantize_level is used.

        """
        import pickle

        np.random.seed(42)

        # Basically what scipy.datasets.ascent() does.
        fname = download_file(
            "https://github.com/scipy/dataset-ascent/blob/main/ascent.dat?raw=true"
        )
        with open(fname, "rb") as f:
            scipy_data = np.array(pickle.load(f))

        data = scipy_data + np.random.randn(512, 512) * 10

        fits.ImageHDU(data).writeto(self.temp("im1.fits"))
        fits.CompImageHDU(
            data,
            compression_type="RICE_1",
            quantize_method=1,
            quantize_level=-1,
            dither_seed=5,
        ).writeto(self.temp("im2.fits"))
        fits.CompImageHDU(
            data,
            compression_type="RICE_1",
            quantize_method=1,
            quantize_level=-100,
            dither_seed=5,
        ).writeto(self.temp("im3.fits"))

        im1 = fits.getdata(self.temp("im1.fits"))
        im2 = fits.getdata(self.temp("im2.fits"))
        im3 = fits.getdata(self.temp("im3.fits"))

        assert not np.array_equal(im2, im3)
        assert np.isclose(np.min(im1 - im2), -0.5, atol=1e-3)
        assert np.isclose(np.max(im1 - im2), 0.5, atol=1e-3)
        assert np.isclose(np.min(im1 - im3), -50, atol=1e-1)
        assert np.isclose(np.max(im1 - im3), 50, atol=1e-1)

    def test_comp_image_hcompression_1_invalid_data(self):
        """
        Tests compression with the HCOMPRESS_1 algorithm with data that is
        not 2D and has a non-2D tile size.
        """

        pytest.raises(
            ValueError,
            fits.CompImageHDU,
            np.zeros((2, 10, 10), dtype=np.float32),
            name="SCI",
            compression_type="HCOMPRESS_1",
            quantize_level=16,
            tile_shape=(2, 10, 10),
        )

    def test_comp_image_hcompress_image_stack(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/171

        Tests that data containing more than two dimensions can be
        compressed with HCOMPRESS_1 so long as the user-supplied tile size can
        be flattened to two dimensions.
        """

        cube = np.arange(300, dtype=np.float32).reshape(3, 10, 10)
        hdu = fits.CompImageHDU(
            data=cube,
            name="SCI",
            compression_type="HCOMPRESS_1",
            quantize_level=16,
            tile_shape=(1, 5, 5),
        )
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as hdul:
            # HCOMPRESSed images are allowed to deviate from the original by
            # about 1/quantize_level of the RMS in each tile.
            assert np.abs(hdul["SCI"].data - cube).max() < 1.0 / 15.0

    def test_subtractive_dither_seed(self):
        """
        Regression test for https://github.com/spacetelescope/PyFITS/issues/32

        Ensure that when floating point data is compressed with the
        SUBTRACTIVE_DITHER_1 quantization method that the correct ZDITHER0 seed
        is added to the header, and that the data can be correctly
        decompressed.
        """

        array = np.arange(100.0).reshape(10, 10)
        csum = (array[0].view("uint8").sum() % 10000) + 1
        hdu = fits.CompImageHDU(
            data=array,
            quantize_method=SUBTRACTIVE_DITHER_1,
            dither_seed=DITHER_SEED_CHECKSUM,
        )
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as hdul:
            assert isinstance(hdul[1], fits.CompImageHDU)
            assert "ZQUANTIZ" in hdul[1]._header
            assert hdul[1]._header["ZQUANTIZ"] == "SUBTRACTIVE_DITHER_1"
            assert "ZDITHER0" in hdul[1]._header
            assert hdul[1]._header["ZDITHER0"] == csum
            assert np.all(hdul[1].data == array)

    def test_disable_image_compression(self):
        with fits.open(self.data("comp.fits"), disable_image_compression=True) as hdul:
            # The compressed image HDU should show up as a BinTableHDU, but
            # *not* a CompImageHDU
            assert isinstance(hdul[1], fits.BinTableHDU)
            assert not isinstance(hdul[1], fits.CompImageHDU)

        with fits.open(self.data("comp.fits")) as hdul:
            assert isinstance(hdul[1], fits.CompImageHDU)

    def test_open_comp_image_in_update_mode(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/167

        Similar to test_open_scaled_in_update_mode(), but specifically for
        compressed images.
        """

        # Copy the original file before making any possible changes to it
        self.copy_file("comp.fits")
        mtime = os.stat(self.temp("comp.fits")).st_mtime

        time.sleep(1)

        fits.open(self.temp("comp.fits"), mode="update").close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert mtime == os.stat(self.temp("comp.fits")).st_mtime

    @pytest.mark.slow
    def test_open_scaled_in_update_mode_compressed(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/88 2

        Identical to test_open_scaled_in_update_mode() but with a compressed
        version of the scaled image.
        """

        # Copy+compress the original file before making any possible changes to
        # it
        with fits.open(self.data("scale.fits"), do_not_scale_image_data=True) as hdul:
            chdu = fits.CompImageHDU(data=hdul[0].data, header=hdul[0].header)
            chdu.writeto(self.temp("scale.fits"))
        mtime = os.stat(self.temp("scale.fits")).st_mtime

        time.sleep(1)

        fits.open(self.temp("scale.fits"), mode="update").close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert mtime == os.stat(self.temp("scale.fits")).st_mtime

        # Insert a slight delay to ensure the mtime does change when the file
        # is changed
        time.sleep(1)

        hdul = fits.open(self.temp("scale.fits"), "update")
        hdul[1].data
        hdul.close()

        # Now the file should be updated with the rescaled data
        assert mtime != os.stat(self.temp("scale.fits")).st_mtime
        hdul = fits.open(self.temp("scale.fits"), mode="update")
        assert hdul[1].data.dtype == np.dtype("float32")
        assert hdul[1].header["BITPIX"] == -32
        assert "BZERO" not in hdul[1].header
        assert "BSCALE" not in hdul[1].header

        # Try reshaping the data, then closing and reopening the file; let's
        # see if all the changes are preserved properly
        hdul[1].data.shape = (42, 10)
        hdul.close()

        hdul = fits.open(self.temp("scale.fits"))
        assert hdul[1].shape == (42, 10)
        assert hdul[1].data.dtype == np.dtype("float32")
        assert hdul[1].header["BITPIX"] == -32
        assert "BZERO" not in hdul[1].header
        assert "BSCALE" not in hdul[1].header
        hdul.close()

    def test_write_comp_hdu_direct_from_existing(self):
        with fits.open(self.data("comp.fits")) as hdul:
            hdul[1].writeto(self.temp("test.fits"))

        with fits.open(self.data("comp.fits")) as hdul1:
            with fits.open(self.temp("test.fits")) as hdul2:
                assert np.all(hdul1[1].data == hdul2[1].data)
                assert comparerecords(
                    hdul1[1].compressed_data, hdul2[1].compressed_data
                )

    def test_rewriting_large_scaled_image_compressed(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/88 1

        Identical to test_rewriting_large_scaled_image() but with a compressed
        image.
        """

        with fits.open(
            self.data("fixed-1890.fits"), do_not_scale_image_data=True
        ) as hdul:
            chdu = fits.CompImageHDU(data=hdul[0].data, header=hdul[0].header)
            chdu.writeto(self.temp("fixed-1890-z.fits"))

        hdul = fits.open(self.temp("fixed-1890-z.fits"))
        orig_data = hdul[1].data
        hdul.writeto(self.temp("test_new.fits"), overwrite=True)
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        assert (hdul[1].data == orig_data).all()
        hdul.close()

        # Just as before, but this time don't touch hdul[0].data before writing
        # back out--this is the case that failed in
        # https://aeon.stsci.edu/ssb/trac/pyfits/ticket/84
        hdul = fits.open(self.temp("fixed-1890-z.fits"))
        hdul.writeto(self.temp("test_new.fits"), overwrite=True)
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        assert (hdul[1].data == orig_data).all()
        hdul.close()

        # Test opening/closing/reopening a scaled file in update mode
        hdul = fits.open(self.temp("fixed-1890-z.fits"), do_not_scale_image_data=True)
        hdul.writeto(
            self.temp("test_new.fits"), overwrite=True, output_verify="silentfix"
        )
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        orig_data = hdul[1].data
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"), mode="update")
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        assert (hdul[1].data == orig_data).all()
        hdul.close()

    def test_scale_back_compressed(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/88 3

        Identical to test_scale_back() but uses a compressed image.
        """

        # Create a compressed version of the scaled image
        with fits.open(self.data("scale.fits"), do_not_scale_image_data=True) as hdul:
            chdu = fits.CompImageHDU(data=hdul[0].data, header=hdul[0].header)
            chdu.writeto(self.temp("scale.fits"))

        with fits.open(self.temp("scale.fits"), mode="update", scale_back=True) as hdul:
            orig_bitpix = hdul[1].header["BITPIX"]
            orig_bzero = hdul[1].header["BZERO"]
            orig_bscale = hdul[1].header["BSCALE"]
            orig_data = hdul[1].data.copy()
            hdul[1].data[0] = 0

        with fits.open(self.temp("scale.fits"), do_not_scale_image_data=True) as hdul:
            assert hdul[1].header["BITPIX"] == orig_bitpix
            assert hdul[1].header["BZERO"] == orig_bzero
            assert hdul[1].header["BSCALE"] == orig_bscale

            zero_point = int(math.floor(-orig_bzero / orig_bscale))
            assert (hdul[1].data[0] == zero_point).all()

        with fits.open(self.temp("scale.fits")) as hdul:
            assert (hdul[1].data[1:] == orig_data[1:]).all()
            # Extra test to ensure that after everything the data is still the
            # same as in the original uncompressed version of the image
            with fits.open(self.data("scale.fits")) as hdul2:
                # Recall we made the same modification to the data in hdul
                # above
                hdul2[0].data[0] = 0
                assert (hdul[1].data == hdul2[0].data).all()

    def test_lossless_gzip_compression(self):
        """Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/198"""

        rng = np.random.default_rng(42)
        noise = rng.normal(size=(20, 20))

        chdu1 = fits.CompImageHDU(data=noise, compression_type="GZIP_1")
        # First make a test image with lossy compression and make sure it
        # wasn't compressed perfectly.  This shouldn't happen ever, but just to
        # make sure the test non-trivial.
        chdu1.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as h:
            assert np.abs(noise - h[1].data).max() > 0.0

        del h

        chdu2 = fits.CompImageHDU(
            data=noise, compression_type="GZIP_1", quantize_level=0.0
        )  # No quantization
        chdu2.writeto(self.temp("test.fits"), overwrite=True)

        with fits.open(self.temp("test.fits")) as h:
            assert (noise == h[1].data).all()

    def test_compression_column_tforms(self):
        """Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/199"""

        # Some interestingly tiled data so that some of it is quantized and
        # some of it ends up just getting gzip-compressed
        data2 = (np.arange(1, 8, dtype=np.float32) * 10)[:, np.newaxis] + np.arange(
            1, 7
        )
        np.random.seed(1337)
        data1 = np.random.uniform(size=(6 * 4, 7 * 4))
        data1[: data2.shape[0], : data2.shape[1]] = data2
        chdu = fits.CompImageHDU(data1, compression_type="RICE_1", tile_shape=(6, 7))
        chdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits"), disable_image_compression=True) as h:
            assert re.match(r"^1PB\(\d+\)$", h[1].header["TFORM1"])
            assert re.match(r"^1PB\(\d+\)$", h[1].header["TFORM2"])

    def test_compression_update_header(self):
        """Regression test for
        https://github.com/spacetelescope/PyFITS/issues/23
        """

        self.copy_file("comp.fits")
        with fits.open(self.temp("comp.fits"), mode="update") as hdul:
            assert isinstance(hdul[1], fits.CompImageHDU)
            hdul[1].header["test1"] = "test"
            hdul[1]._header["test2"] = "test2"

        with fits.open(self.temp("comp.fits")) as hdul:
            assert "test1" in hdul[1].header
            assert hdul[1].header["test1"] == "test"
            assert "test2" in hdul[1].header
            assert hdul[1].header["test2"] == "test2"

        # Test update via index now:
        with fits.open(self.temp("comp.fits"), mode="update") as hdul:
            hdr = hdul[1].header
            hdr[hdr.index("TEST1")] = "foo"

        with fits.open(self.temp("comp.fits")) as hdul:
            assert hdul[1].header["TEST1"] == "foo"

        # Test slice updates
        with fits.open(self.temp("comp.fits"), mode="update") as hdul:
            hdul[1].header["TEST*"] = "qux"

        with fits.open(self.temp("comp.fits")) as hdul:
            assert list(hdul[1].header["TEST*"].values()) == ["qux", "qux"]

        with fits.open(self.temp("comp.fits"), mode="update") as hdul:
            hdr = hdul[1].header
            idx = hdr.index("TEST1")
            hdr[idx : idx + 2] = "bar"

        with fits.open(self.temp("comp.fits")) as hdul:
            assert list(hdul[1].header["TEST*"].values()) == ["bar", "bar"]

        # Test updating a specific COMMENT card duplicate
        with fits.open(self.temp("comp.fits"), mode="update") as hdul:
            hdul[1].header[("COMMENT", 1)] = "I am fire. I am death!"

        with fits.open(self.temp("comp.fits")) as hdul:
            assert hdul[1].header["COMMENT"][1] == "I am fire. I am death!"
            assert hdul[1]._header["COMMENT"][1] == "I am fire. I am death!"

        # Test deleting by keyword and by slice
        with fits.open(self.temp("comp.fits"), mode="update") as hdul:
            hdr = hdul[1].header
            del hdr["COMMENT"]
            idx = hdr.index("TEST1")
            del hdr[idx : idx + 2]

        with fits.open(self.temp("comp.fits")) as hdul:
            assert "COMMENT" not in hdul[1].header
            assert "COMMENT" not in hdul[1]._header
            assert "TEST1" not in hdul[1].header
            assert "TEST1" not in hdul[1]._header
            assert "TEST2" not in hdul[1].header
            assert "TEST2" not in hdul[1]._header

    def test_compression_update_header_with_reserved(self):
        """
        Ensure that setting reserved keywords related to the table data
        structure on CompImageHDU image headers fails.
        """

        def test_set_keyword(hdr, keyword, value):
            with pytest.warns(UserWarning) as w:
                hdr[keyword] = value
            assert len(w) == 1
            assert str(w[0].message).startswith(f"Keyword {keyword!r} is reserved")
            assert keyword not in hdr

        with fits.open(self.data("comp.fits")) as hdul:
            hdr = hdul[1].header
            test_set_keyword(hdr, "TFIELDS", 8)
            test_set_keyword(hdr, "TTYPE1", "Foo")
            test_set_keyword(hdr, "ZCMPTYPE", "ASDF")
            test_set_keyword(hdr, "ZVAL1", "Foo")

    def test_compression_header_append(self):
        with fits.open(self.data("comp.fits")) as hdul:
            imghdr = hdul[1].header
            tblhdr = hdul[1]._header
            with pytest.warns(UserWarning, match="Keyword 'TFIELDS' is reserved") as w:
                imghdr.append("TFIELDS")
            assert len(w) == 1
            assert "TFIELDS" not in imghdr

            imghdr.append(("FOO", "bar", "qux"), end=True)
            assert "FOO" in imghdr
            assert imghdr[-1] == "bar"
            assert "FOO" in tblhdr
            assert tblhdr[-1] == "bar"

            imghdr.append(("CHECKSUM", "abcd1234"))
            assert "CHECKSUM" in imghdr
            assert imghdr["CHECKSUM"] == "abcd1234"
            assert "CHECKSUM" not in tblhdr
            assert "ZHECKSUM" in tblhdr
            assert tblhdr["ZHECKSUM"] == "abcd1234"

    def test_compression_header_append2(self):
        """
        Regression test for issue https://github.com/astropy/astropy/issues/5827
        """
        with fits.open(self.data("comp.fits")) as hdul:
            header = hdul[1].header
            while len(header) < 1000:
                header.append()  # pad with grow room

            # Append stats to header:
            header.append(("Q1_OSAVG", 1, "[adu] quadrant 1 overscan mean"))
            header.append(("Q1_OSSTD", 1, "[adu] quadrant 1 overscan stddev"))
            header.append(("Q1_OSMED", 1, "[adu] quadrant 1 overscan median"))

    def test_compression_header_insert(self):
        with fits.open(self.data("comp.fits")) as hdul:
            imghdr = hdul[1].header
            tblhdr = hdul[1]._header
            # First try inserting a restricted keyword
            with pytest.warns(UserWarning, match="Keyword 'TFIELDS' is reserved") as w:
                imghdr.insert(1000, "TFIELDS")
            assert len(w) == 1
            assert "TFIELDS" not in imghdr
            assert tblhdr.count("TFIELDS") == 1

            # First try keyword-relative insert
            imghdr.insert("TELESCOP", ("OBSERVER", "Phil Plait"))
            assert "OBSERVER" in imghdr
            assert imghdr.index("OBSERVER") == imghdr.index("TELESCOP") - 1
            assert "OBSERVER" in tblhdr
            assert tblhdr.index("OBSERVER") == tblhdr.index("TELESCOP") - 1

            # Next let's see if an index-relative insert winds up being
            # sensible
            idx = imghdr.index("OBSERVER")
            imghdr.insert("OBSERVER", ("FOO",))
            assert "FOO" in imghdr
            assert imghdr.index("FOO") == idx
            assert "FOO" in tblhdr
            assert tblhdr.index("FOO") == tblhdr.index("OBSERVER") - 1

    def test_compression_header_set_before_after(self):
        with fits.open(self.data("comp.fits")) as hdul:
            imghdr = hdul[1].header
            tblhdr = hdul[1]._header

            with pytest.warns(UserWarning, match="Keyword 'ZBITPIX' is reserved ") as w:
                imghdr.set("ZBITPIX", 77, "asdf", after="XTENSION")
            assert len(w) == 1
            assert "ZBITPIX" not in imghdr
            assert tblhdr.count("ZBITPIX") == 1
            assert tblhdr["ZBITPIX"] != 77

            # Move GCOUNT before PCOUNT (not that there's any reason you'd
            # *want* to do that, but it's just a test...)
            imghdr.set("GCOUNT", 99, before="PCOUNT")
            assert imghdr.index("GCOUNT") == imghdr.index("PCOUNT") - 1
            assert imghdr["GCOUNT"] == 99
            assert tblhdr.index("ZGCOUNT") == tblhdr.index("ZPCOUNT") - 1
            assert tblhdr["ZGCOUNT"] == 99
            assert tblhdr.index("PCOUNT") == 5
            assert tblhdr.index("GCOUNT") == 6
            assert tblhdr["GCOUNT"] == 1

            imghdr.set("GCOUNT", 2, after="PCOUNT")
            assert imghdr.index("GCOUNT") == imghdr.index("PCOUNT") + 1
            assert imghdr["GCOUNT"] == 2
            assert tblhdr.index("ZGCOUNT") == tblhdr.index("ZPCOUNT") + 1
            assert tblhdr["ZGCOUNT"] == 2
            assert tblhdr.index("PCOUNT") == 5
            assert tblhdr.index("GCOUNT") == 6
            assert tblhdr["GCOUNT"] == 1

    def test_compression_header_append_commentary(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2363
        """

        hdu = fits.CompImageHDU(np.array([0], dtype=np.int32))
        hdu.header["COMMENT"] = "hello world"
        assert hdu.header["COMMENT"] == ["hello world"]
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as hdul:
            assert hdul[1].header["COMMENT"] == ["hello world"]

    def test_compression_with_gzip_column(self):
        """
        Regression test for https://github.com/spacetelescope/PyFITS/issues/71
        """

        arr = np.zeros((2, 7000), dtype="float32")

        # The first row (which will be the first compressed tile) has a very
        # wide range of values that will be difficult to quantize, and should
        # result in use of a GZIP_COMPRESSED_DATA column
        arr[0] = np.linspace(0, 1, 7000)
        arr[1] = np.random.normal(size=7000)

        hdu = fits.CompImageHDU(data=arr)
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as hdul:
            comp_hdu = hdul[1]

            # GZIP-compressed tile should compare exactly
            assert np.all(comp_hdu.data[0] == arr[0])
            # The second tile uses lossy compression and may be somewhat off,
            # so we don't bother comparing it exactly

    def test_duplicate_compression_header_keywords(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2750

        Tests that the fake header (for the compressed image) can still be read
        even if the real header contained a duplicate ZTENSION keyword (the
        issue applies to any keyword specific to the compression convention,
        however).
        """

        arr = np.arange(100, dtype=np.int32)
        hdu = fits.CompImageHDU(data=arr)

        header = hdu._header
        # append the duplicate keyword
        hdu._header.append(("ZTENSION", "IMAGE"))
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as hdul:
            assert header == hdul[1]._header
            # There's no good reason to have a duplicate keyword, but
            # technically it isn't invalid either :/
            assert hdul[1]._header.count("ZTENSION") == 2

    def test_scale_bzero_with_compressed_int_data(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/4600
        and https://github.com/astropy/astropy/issues/4588

        Identical to test_scale_bzero_with_int_data() but uses a compressed
        image.
        """

        a = np.arange(100, 200, dtype=np.int16)

        hdu1 = fits.CompImageHDU(data=a.copy())
        hdu2 = fits.CompImageHDU(data=a.copy())
        # Previously the following line would throw a TypeError,
        # now it should be identical to the integer bzero case
        hdu1.scale("int16", bzero=99.0)
        hdu2.scale("int16", bzero=99)
        assert np.allclose(hdu1.data, hdu2.data)

    def test_scale_back_compressed_uint_assignment(self):
        """
        Extend fix for #4600 to assignment to data

        Identical to test_scale_back_uint_assignment() but uses a compressed
        image.

        Suggested by:
        https://github.com/astropy/astropy/pull/4602#issuecomment-208713748
        """

        a = np.arange(100, 200, dtype=np.uint16)
        fits.CompImageHDU(a).writeto(self.temp("test.fits"))
        with fits.open(self.temp("test.fits"), mode="update", scale_back=True) as hdul:
            hdul[1].data[:] = 0
            assert np.allclose(hdul[1].data, 0)

    def test_compressed_header_missing_znaxis(self):
        a = np.arange(100, 200, dtype=np.uint16)
        comp_hdu = fits.CompImageHDU(a)
        comp_hdu._header.pop("ZNAXIS")
        with pytest.raises(KeyError):
            comp_hdu.compressed_data
        comp_hdu = fits.CompImageHDU(a)
        comp_hdu._header.pop("ZBITPIX")
        with pytest.raises(KeyError):
            comp_hdu.compressed_data

    def test_compressed_header_double_extname(self):
        """Test that a double EXTNAME with one default value does not
        mask the non-default value."""
        with fits.open(self.data("double_ext.fits")) as hdul:
            hdu = hdul[1]

            # Raw header has 2 EXTNAME entries
            indices = hdu._header._keyword_indices["EXTNAME"]
            assert len(indices) == 2

            # The non-default name should be returned.
            assert hdu.name == "ccd00"
            assert "EXTNAME" in hdu.header
            assert hdu.name == hdu.header["EXTNAME"]

            # There should be 1 non-default EXTNAME entries.
            indices = hdu.header._keyword_indices["EXTNAME"]
            assert len(indices) == 1

            # Test header sync from property set.
            new_name = "NEW_NAME"
            hdu.name = new_name
            assert hdu.name == new_name
            assert hdu.header["EXTNAME"] == new_name
            assert hdu._header["EXTNAME"] == new_name
            assert hdu._image_header["EXTNAME"] == new_name

            # Check that setting the header will change the name property.
            hdu.header["EXTNAME"] = "NEW2"
            assert hdu.name == "NEW2"

            hdul.writeto(self.temp("tmp.fits"), overwrite=True)
            with fits.open(self.temp("tmp.fits")) as hdul1:
                hdu1 = hdul1[1]
                assert len(hdu1._header._keyword_indices["EXTNAME"]) == 1
                assert hdu1.name == "NEW2"

            # Check that deleting EXTNAME will and setting the name will
            # work properly.
            del hdu.header["EXTNAME"]
            hdu.name = "RE-ADDED"
            assert hdu.name == "RE-ADDED"

            with pytest.raises(TypeError):
                hdu.name = 42

    def test_compressed_header_extname(self):
        """Test consistent EXTNAME / hdu name interaction."""
        name = "FOO"
        hdu = fits.CompImageHDU(data=np.arange(10), name=name)
        assert hdu._header["EXTNAME"] == name
        assert hdu.header["EXTNAME"] == name
        assert hdu.name == name

        name = "BAR"
        hdu.name = name
        assert hdu._header["EXTNAME"] == name
        assert hdu.header["EXTNAME"] == name
        assert hdu.name == name

        assert len(hdu._header._keyword_indices["EXTNAME"]) == 1

    def test_compressed_header_minimal(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/11694

        Tests that CompImageHDU can be initialized with a Header that
        contains few or no cards, and doesn't require specific cards
        such as 'BITPIX' or 'NAXIS'.
        """
        fits.CompImageHDU(data=np.arange(10), header=fits.Header())
        header = fits.Header({"HELLO": "world"})
        hdu = fits.CompImageHDU(data=np.arange(10), header=header)
        assert hdu.header["HELLO"] == "world"

    @pytest.mark.parametrize(
        ("keyword", "dtype", "expected"),
        [
            ("BSCALE", np.uint8, np.float32),
            ("BSCALE", np.int16, np.float32),
            ("BSCALE", np.int32, np.float64),
            ("BZERO", np.uint8, np.float32),
            ("BZERO", np.int16, np.float32),
            ("BZERO", np.int32, np.float64),
        ],
    )
    def test_compressed_scaled_float(self, keyword, dtype, expected):
        """
        If BSCALE,BZERO is set to floating point values, the image
        should be floating-point.

        https://github.com/astropy/astropy/pull/6492

        Parameters
        ----------
        keyword : `str`
            Keyword to set to a floating-point value to trigger
            floating-point pixels.
        dtype : `numpy.dtype`
            Type of original array.
        expected : `numpy.dtype`
            Expected type of uncompressed array.
        """
        value = 1.23345  # A floating-point value
        hdu = fits.CompImageHDU(np.arange(0, 10, dtype=dtype))
        hdu.header[keyword] = value
        hdu.writeto(self.temp("test.fits"))
        del hdu
        with fits.open(self.temp("test.fits")) as hdu:
            assert hdu[1].header[keyword] == value
            assert hdu[1].data.dtype == expected

    @pytest.mark.parametrize(
        "dtype", (np.uint8, np.int16, np.uint16, np.int32, np.uint32)
    )
    def test_compressed_integers(self, dtype):
        """Test that the various integer dtypes are correctly written and read.

        Regression test for https://github.com/astropy/astropy/issues/9072

        """
        mid = np.iinfo(dtype).max // 2
        data = np.arange(mid - 50, mid + 50, dtype=dtype)
        testfile = self.temp("test.fits")
        hdu = fits.CompImageHDU(data=data)
        hdu.writeto(testfile, overwrite=True)
        new = fits.getdata(testfile)
        np.testing.assert_array_equal(data, new)

    @pytest.mark.parametrize(
        ("dtype", "compression_type"), product(("f", "i4"), COMPRESSION_TYPES)
    )
    def test_write_non_contiguous_data(self, dtype, compression_type):
        """
        Regression test for https://github.com/astropy/astropy/issues/2150

        This used to require changing the whole array to be C-contiguous before
        passing to CFITSIO, but we no longer need this - our explicit conversion
        to bytes in the compression codecs returns contiguous bytes for each
        tile on-the-fly.
        """

        orig = np.arange(400, dtype=dtype).reshape((20, 20), order="f")[::2, ::2]
        assert not orig.flags.contiguous
        primary = fits.PrimaryHDU()
        hdu = fits.CompImageHDU(orig, compression_type=compression_type)
        hdulist = fits.HDUList([primary, hdu])
        hdulist.writeto(self.temp("test.fits"))

        actual = fits.getdata(self.temp("test.fits"))
        assert_equal(orig, actual)

    def test_slice_and_write_comp_hdu(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/9955
        """
        with fits.open(self.data("comp.fits")) as hdul:
            hdul[1].data = hdul[1].data[:200, :100]
            assert not hdul[1].data.flags.contiguous
            hdul[1].writeto(self.temp("test.fits"))

        with fits.open(self.data("comp.fits")) as hdul1:
            with fits.open(self.temp("test.fits")) as hdul2:
                assert_equal(hdul1[1].data[:200, :100], hdul2[1].data)

    def test_comp_image_deprecated_tile_size(self):
        # Ensure that tile_size works but is deprecated. This test
        # can be removed once support for tile_size is removed.

        with pytest.warns(
            AstropyDeprecationWarning,
            match="The tile_size argument has been deprecated",
        ):
            chdu = fits.CompImageHDU(np.zeros((3, 4, 5)), tile_size=(5, 2, 1))

        assert chdu.tile_shape == (1, 2, 5)

    def test_comp_image_deprecated_tile_size_and_tile_shape(self):
        # Make sure that tile_size and tile_shape are not both specified

        with pytest.warns(AstropyDeprecationWarning) as w:
            with pytest.raises(
                ValueError, match="Cannot specify both tile_size and tile_shape."
            ):
                fits.CompImageHDU(
                    np.zeros((3, 4, 5)), tile_size=(5, 2, 1), tile_shape=(3, 2, 3)
                )

    def test_comp_image_properties_default(self):
        chdu = fits.CompImageHDU(np.zeros((3, 4, 5)))
        assert chdu.tile_shape == (1, 1, 5)
        assert chdu.compression_type == "RICE_1"

    def test_comp_image_properties_set(self):
        chdu = fits.CompImageHDU(
            np.zeros((3, 4, 5)), compression_type="PLIO_1", tile_shape=(2, 3, 4)
        )
        assert chdu.tile_shape == (2, 3, 4)
        assert chdu.compression_type == "PLIO_1"

    def test_compressed_optional_prefix_tform(self, tmp_path):
        # Regression test for a bug that caused an error if a
        # compressed file had TFORM missing the optional 1 prefix

        data = np.zeros((3, 4, 5))

        hdu1 = fits.CompImageHDU(data=data)
        hdu1.writeto(tmp_path / "compressed.fits")

        with fits.open(
            tmp_path / "compressed.fits", disable_image_compression=True, mode="update"
        ) as hdul:
            assert hdul[1].header["TFORM1"] == "1PB(0)"
            assert hdul[1].header["TFORM2"] == "1PB(24)"
            hdul[1].header["TFORM1"] = "PB(0)"
            hdul[1].header["TFORM2"] = "PB(24)"

        with fits.open(
            tmp_path / "compressed.fits", disable_image_compression=True
        ) as hdul:
            assert hdul[1].header["TFORM1"] == "PB(0)"
            assert hdul[1].header["TFORM2"] == "PB(24)"

        with fits.open(tmp_path / "compressed.fits") as hdul:
            assert_equal(hdul[1].data, data)


class TestCompHDUSections:
    @pytest.fixture(autouse=True)
    def setup_method(self, tmp_path):
        shape = (13, 17, 25)
        self.data = np.arange(np.prod(shape)).reshape(shape).astype(np.int32)

        header1 = fits.Header()
        hdu1 = fits.CompImageHDU(
            self.data, header1, compression_type="RICE_1", tile_shape=(5, 4, 5)
        )

        header2 = fits.Header()
        header2["BSCALE"] = 2
        header2["BZERO"] = 100
        hdu2 = fits.CompImageHDU(
            self.data, header2, compression_type="RICE_1", tile_shape=(5, 4, 5)
        )
        hdulist = fits.HDUList([fits.PrimaryHDU(), hdu1, hdu2])
        hdulist.writeto(tmp_path / "sections.fits")

        self.hdul = fits.open(tmp_path / "sections.fits")

    def teardown_method(self):
        self.hdul.close()
        self.hdul = None

    @given(basic_indices((13, 17, 25)))
    def test_section_slicing(self, index):
        assert_equal(self.hdul[1].section[index], self.hdul[1].data[index])
        assert_equal(self.hdul[1].section[index], self.data[index])

    @given(basic_indices((13, 17, 25)))
    def test_section_slicing_scaling(self, index):
        assert_equal(self.hdul[2].section[index], self.hdul[2].data[index])
        assert_equal(self.hdul[2].section[index], self.data[index] * 2 + 100)


def test_comphdu_fileobj():
    # Regression test for a bug that caused an error to happen
    # internally when reading the data if requested data shapes
    # were not plain integers - this was triggered when accessing
    # sections on data backed by certain kinds of objects such as
    # BytesIO (but not regular file handles)

    data = np.arange(6).reshape((2, 3)).astype(np.int32)

    byte_buffer = BytesIO()

    header = fits.Header()
    hdu = fits.CompImageHDU(data, header, compression_type="RICE_1")
    hdu.writeto(byte_buffer)

    byte_buffer.seek(0)

    hdu2 = fits.open(byte_buffer, mode="readonly")[1]
    assert hdu2.section[1, 2] == 5


def test_comphdu_bscale(tmp_path):
    """
    Regression test for a bug that caused extensions that used BZERO and BSCALE
    that got turned into CompImageHDU to end up with BZERO/BSCALE before the
    TFIELDS.
    """

    filename1 = tmp_path / "3hdus.fits"
    filename2 = tmp_path / "3hdus_comp.fits"

    x = np.random.random((100, 100)) * 100

    x0 = fits.PrimaryHDU()
    x1 = fits.ImageHDU(np.array(x - 50, dtype=int), uint=True)
    x1.header["BZERO"] = 20331
    x1.header["BSCALE"] = 2.3
    hdus = fits.HDUList([x0, x1])
    hdus.writeto(filename1)

    # fitsverify (based on cfitsio) should fail on this file, only seeing the
    # first HDU.
    with fits.open(filename1) as hdus:
        hdus[1] = fits.CompImageHDU(
            data=hdus[1].data.astype(np.uint32), header=hdus[1].header
        )
        hdus.writeto(filename2)

    # open again and verify
    with fits.open(filename2) as hdus:
        hdus[1].verify("exception")


def test_image_write_readonly(tmp_path):
    # Regression test to make sure that we can write out read-only arrays (#5512)

    x = np.array([1.0, 2.0, 3.0])
    x.setflags(write=False)
    ghdu = fits.CompImageHDU(data=x)
    # add_datasum does not work for CompImageHDU
    # ghdu.add_datasum()

    filename = tmp_path / "test2.fits"

    ghdu.writeto(filename)

    with fits.open(filename) as hdulist:
        assert_equal(hdulist[1].data, [1.0, 2.0, 3.0])
